from typing import Callable, Optional, Tuple
import picwriter.components as pc

from pp.components.waveguide_template import wg_strip
from pp.picwriter2component import picwriter2component
import pp
from pp.component import Component


@pp.autoname
def dbr2(
    length: float = 10.0,
    period: float = 0.85,
    dc: float = 0.5,
    w1: float = 0.4,
    w2: float = 1.0,
    taper_length: float = 20.0,
    fins: bool = False,
    fin_size: Tuple[float, float] = (0.2, 0.05),
    port: Tuple[int, int] = (0, 0),
    direction: str = "EAST",
    waveguide_template: Callable = wg_strip,
    waveguide_template_dbr: Optional[Callable] = None,
    **kwargs
) -> Component:
    """ Distributed Bragg Reflector Cell class.  Tapers the input waveguide to a periodic waveguide structure with varying width (1-D photonic crystal).

    Args:
       wgt (WaveguideTemplate):  WaveguideTemplate object
       length (float): Length of the DBR region.
       period (float): Period of the repeated unit.
       dc (float): Duty cycle of the repeated unit (must be a float between 0 and 1.0).
       w1 (float): Width of the thin section of the waveguide.  w1 = 0 corresponds to disconnected periodic blocks.
       w2 (float): Width of the wide section of the waveguide
       taper_length (float): Length of the taper between the input/output waveguide and the DBR region.  Defaults to 20.0.
       fins (boolean): If `True`, adds fins to the input/output waveguides.  In this case a different template for the component must be specified.  This feature is useful when performing electron-beam lithography and using different beam currents for fine features (helps to reduce stitching errors).  Defaults to `False`
       fin_size ((x,y) Tuple): Specifies the x- and y-size of the `fins`.  Defaults to 200 nm x 50 nm
       dbr_wgt (WaveguideTemplate): If `fins` above is True, a WaveguideTemplate (dbr_wgt) must be specified.  This defines the layertype / datatype of the DBR (which will be separate from the input/output waveguides).  Defaults to `None`
       port (tuple): Cartesian coordinate of the input port.  Defaults to (0,0).
       direction (string): Direction that the component will point *towards*, can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`, OR an angle (float, in radians)
       waveguide_template (WaveguideTemplate): Picwriter WaveguideTemplate object
       wg_width: 0.5
       wg_layer: pp.LAYER.WG[0]
       wg_datatype: pp.LAYER.WG[1]
       clad_layer: pp.LAYER.WGCLAD[0]
       clad_datatype: pp.LAYER.WGCLAD[1]
       bend_radius: 10
       cladding_offset: 3

    .. code::

                 period
        <-----><-------->
                _________
        _______|

          w1       w2       ...  n times
        _______
               |_________



    .. plot::
      :include-source:

      import pp

      c = pp.c.dbr2(length=10, period=0.85, dc=0.5, w2=1, w1=0.4)
      pp.plotgds(c)

    """

    waveguide_template_dbr = waveguide_template_dbr or waveguide_template(wg_width=w2)

    c = pc.DBR(
        wgt=pp.call_if_func(waveguide_template, wg_width=w2, **kwargs),
        length=length,
        period=period,
        dc=dc,
        w_phc=w1,
        taper_length=taper_length,
        fins=fins,
        fin_size=fin_size,
        dbr_wgt=waveguide_template_dbr,
        port=port,
        direction=direction,
    )

    return picwriter2component(c)


if __name__ == "__main__":
    import pp

    c = dbr2(length=10, period=0.85, dc=0.5, w2=1, w1=0.4, taper_length=20, fins=True)
    pp.show(c)



""" define names, clean names and values
"""
import uuid
import functools
from inspect import signature
import hashlib
from typing import Any, Callable
import numpy as np
from phidl import Device
from pp.add_pins import add_pins_and_outline

MAX_NAME_LENGTH = 32

NAME_TO_DEVICE = {}


def join_first_letters(name: str) -> str:
    """ join the first letter of a name separated with underscores (taper_length -> TL) """
    return "".join([x[0] for x in name.split("_") if x])


component_type_to_name = dict(import_phidl_component="phidl")


def get_component_name(component_type: str, **kwargs) -> str:
    name = component_type
    for k, v in component_type_to_name.items():
        name = name.replace(k, v)
    if kwargs:
        name += "_" + dict2name(**kwargs)
    return name


def autoname(component_function: Callable) -> Callable:
    """ decorator for auto-naming component functions
    if no Keyword argument `name`  is passed it creates a name by concenating all Keyword arguments

    To avoid that 2 exact cells are not references of the same cell autoname has a cache where if component has already been build it will return the component from the cache
    You can always over-ride this with `cache = False`
    This is helpful when you are changing the code inside the function that is being cached.

    Args:
        name (str):
        cache (bool): caches functions with same name
        uid (bool): adds a unique id to the name
        pins (bool): add pins
        pins_function (function): function to add pins

    .. plot::
      :include-source:

      import pp

      @pp.autoname
      def rectangle(size=(4,2), layer=0):
          c = pp.Component()
          w, h = size
          points = [[w, h], [w, 0], [0, 0], [0, h]]
          c.add_polygon(points, layer=layer)
          return c

      c = rectangle(layer=1)
      c << pp.c.text(text=c.name, size=1)
      pp.plotgds(c)

    """

    @functools.wraps(component_function)
    def _autoname(*args, **kwargs):
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
        arguments = ", ".join(args_repr + kwargs_repr)

        if args:
            raise ValueError(
                f"autoname supports only Keyword args for `{component_function.__name__}({arguments})`"
            )
        cache = kwargs.pop("cache", True)
        uid = kwargs.pop("uid", False)
        pins = kwargs.pop("pins", False)
        pins_function = kwargs.pop("pins_function", add_pins_and_outline)

        component_type = component_function.__name__
        name = kwargs.pop("name", get_component_name(component_type, **kwargs),)

        if uid:
            name += f"_{str(uuid.uuid4())[:8]}"

        kwargs.pop("ignore_from_name", [])
        sig = signature(component_function)
        # assert_first_letters_are_different(**sig.parameters)

        if "args" not in sig.parameters and "kwargs" not in sig.parameters:
            for key in kwargs.keys():
                assert (
                    key in sig.parameters.keys()
                ), f"`{key}` key not in {list(sig.parameters.keys())} for {component_type}"

        if cache and name in NAME_TO_DEVICE:
            return NAME_TO_DEVICE[name]
        else:
            component = component_function(**kwargs)
            component.name = name
            component.module = component_function.__module__
            component.function_name = component_function.__name__

            if len(name) > MAX_NAME_LENGTH:
                component.name_long = name
                component.name = (
                    f"{component_type}_{hashlib.md5(name.encode()).hexdigest()[:8]}"
                )

            if not hasattr(component, "settings"):
                component.settings = {}
            component.settings.update(
                **{
                    p.name: p.default
                    for p in sig.parameters.values()
                    if not callable(p.default)
                }
            )
            component.settings.update(**kwargs)
            component.settings_changed = kwargs.copy()
            if pins:
                pins_function(component)
            NAME_TO_DEVICE[name] = component
            return component

    return _autoname


def dict2hash(**kwargs) -> str:
    ignore_from_name = kwargs.pop("ignore_from_name", [])
    h = hashlib.sha256()
    for key in sorted(kwargs):
        if key not in ignore_from_name:
            value = kwargs[key]
            value = clean_value(value)
            h.update(f"{key}{value}".encode())
    return h.hexdigest()


def dict2name(prefix: None = None, **kwargs) -> str:
    """ returns name from a dict """
    ignore_from_name = kwargs.pop("ignore_from_name", [])

    if prefix:
        label = [prefix]
    else:
        label = []
    for key in sorted(kwargs):
        if key not in ignore_from_name:
            value = kwargs[key]
            key = join_first_letters(key)
            value = clean_value(value)
            label += [f"{key.upper()}{value}"]
    label = "_".join(label)
    return clean_name(label)


def assert_first_letters_are_different(**kwargs):
    """ avoids having name colissions of different args with the same first letter """
    first_letters = [join_first_letters(k) for k in kwargs.keys() if k != "layer"]
    assert len(set(first_letters)) == len(
        first_letters
    ), f"Possible Duplicated name because {kwargs.keys()} has repeated first letters {first_letters}"
    if not len(set(first_letters)) == len(first_letters):
        print(
            f"Possible Duplicated name because {kwargs.keys()} has repeated first letters {first_letters}"
        )


def clean_dict(d):
    """ cleans dictionary keys
    """
    for k, v in d.items():
        if isinstance(v, dict):
            clean_dict(v)
        elif callable(v):
            d[k] = v.__name__


def clean_list(d):
    for v in d:
        if isinstance(v, list):
            clean_list(v)
        elif callable(v):
            d.replace(v, v.__name__)


def clean_name(name: str) -> str:
    """ Ensures that gds cells are composed of [a-zA-Z0-9]

    FIXME: only a few characters are currently replaced.
        This function has been updated only on case-by-case basis
    """
    replace_map = {
        " ": "_",
        "!": "_",
        "#": "_",
        "%": "_",
        "(": "",
        ")": "",
        "*": "_",
        ",": "_",
        "-": "m",
        ".": "p",
        "/": "_",
        ":": "_",
        "=": "",
        "@": "_",
        "[": "",
        "]": "",
    }
    for k, v in list(replace_map.items()):
        name = name.replace(k, v)
    return name


def clean_value(value: Any) -> str:
    """ returns more readable value (integer)
    if number is < 1:
        returns number units in nm (integer)
    """

    if isinstance(value, int):
        value = str(value)
    elif isinstance(value, (float, np.float64)):
        if 1e9 > value > 1e12:
            value = f"{int(value/1e9)}G"
        elif 1e6 > value > 1e9:
            value = f"{int(value/1e6)}M"
        elif 1e3 > value > 1e6:
            value = f"{int(value/1e3)}K"
        elif 1 > value > 1e-3:
            value = f"{int(value*1e3)}m"
        elif 1e-6 < value < 1e-3:
            value = f"{int(value*1e6)}u"
        elif 1e-9 < value < 1e-6:
            value = f"{int(value*1e9)}n"
        elif 1e-12 < value < 1e-9:
            value = f"{int(value*1e12)}p"
        else:
            value = f"{value:.2f}"
    elif isinstance(value, list):
        value = "_".join(clean_value(v) for v in value)
    elif isinstance(value, tuple):
        value = "_".join(clean_value(v) for v in value)
    elif isinstance(value, dict):
        value = dict2name(**value)
    elif isinstance(value, Device):
        value = clean_name(value.name)
    elif callable(value):
        value = value.__name__
    else:
        value = clean_name(str(value))
    return value


class _Dummy:
    pass


@autoname
def _dummy(length=3, wg_width=0.5):
    c = _Dummy()
    c.name = ""
    c.settings = {}
    return c


def test_autoname():
    name_base = _dummy().name
    assert name_base == "_dummy"
    name_int = _dummy(length=3).name
    assert name_int == "_dummy_L3"
    name_float = _dummy(wg_width=0.5).name
    # assert name_float == "_dummy_WW500m"
    name_length_first = _dummy(length=3, wg_width=0.5).name
    name_width_first = _dummy(wg_width=0.5, length=3).name
    assert name_length_first == name_width_first

    name_float = _dummy(wg_width=0.5).name
    # assert name_float == "_dummy_WW0p5"
    print(name_float)


def test_clean_value():
    assert clean_value(0.5) == "500m"
    assert clean_value(5) == "5"


def test_clean_name():
    assert clean_name("wg(:_=_2852") == "wg___2852"


if __name__ == "__main__":
    # test_autoname()
    import pp

    # print(clean_value(pp.c.waveguide))

    # c = pp.c.waveguide(polarization="TMeraer")
    # print(c.get_settings()["polarization"])

    c = pp.c.waveguide(length=11)
    print(c)
    # print(c)
    # pp.show(c)

    # print(clean_name("Waveguidenol1_(:_=_2852"))
    # print(clean_value(1.2))
    # print(clean_value(0.2))
    # print(clean_value([1, [2.4324324, 3]]))
    # print(clean_value([1, 2.4324324, 3]))
    # print(clean_value((0.001, 24)))
    # print(clean_value({"a": 1, "b": 2}))



import pp
from pp.components import bend_circular
from pp.components import waveguide
from pp.components.component_sequence import component_sequence
from pp.routing.add_fiber_array import add_fiber_array


def _get_bend_size(bend90):
    p1, p2 = list(bend90.ports.values())[:2]
    bsx = abs(p2.x - p1.x)
    bsy = abs(p2.y - p1.y)
    return max(bsx, bsy)


@pp.autoname
def cutback_bend(bend90, straight_length=5.0, n_steps=6, n_stairs=5):
    """ Deprecated! use cutback_bend90 instead!
    this is a stair
        _
      _|
    _|

    _: this is a step

    """

    wg = waveguide(length=straight_length, width=bend90.ports["W0"].width)
    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "A": (bend90, "W0", "N0"),
        "B": (bend90, "N0", "W0"),
        "S": (wg, "W0", "E0"),
    }

    # Generate the sequence of staircases
    s = ""
    for i in range(n_stairs):
        s += "ASBS" * n_steps
        s += "ASAS" if i % 2 == 0 else "BSBS"
    s = s[:-4]

    # Create the component from the sequence
    c = component_sequence(s, string_to_device_in_out_ports, start_orientation=90)
    c.update_settings(n_bends=n_steps * n_stairs * 2 + n_stairs * 2 - 2)
    return c


@pp.autoname
def cutback_bend90(
    bend90=pp.c.bend_euler90,
    straight_length=5.0,
    n_steps=6,
    cols=6,
    spacing=5,
    wg_loop_length=None,
    waveguide_factory=waveguide,
):
    """
       _
    |_| |

    """
    bend90 = pp.call_if_func(bend90)

    wg = waveguide_factory(length=straight_length, width=bend90.ports["W0"].width)
    if wg_loop_length is None:
        wg_loop_length = 2 * _get_bend_size(bend90) + spacing + straight_length

    wg_loop = waveguide_factory(length=wg_loop_length, width=bend90.ports["W0"].width,)
    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "A": (bend90, "W0", "N0"),
        "B": (bend90, "N0", "W0"),
        "-": (wg, "W0", "E0"),
        "|": (wg_loop, "W0", "E0"),
    }

    # Generate the sequence of staircases
    s = ""
    for i in range(cols):
        if i % 2 == 0:  # even row
            s += "A-A-B-B-" * n_steps + "|"
        else:
            s += "B-B-A-A-" * n_steps + "|"
    s = s[:-1]

    # Create the component from the sequence
    c = component_sequence(s, string_to_device_in_out_ports, start_orientation=0)
    c.update_settings(n_bends=n_steps * cols * 4)
    return c


def staircase(
    bend90=pp.c.bend_euler90,
    length_v=5.0,
    length_h=5.0,
    n_steps=4,
    waveguide_factory=waveguide,
):
    bend90 = pp.call_if_func(bend90)

    wgh = waveguide_factory(length=length_h, width=bend90.ports["W0"].width)
    wgv = waveguide_factory(length=length_v, width=bend90.ports["W0"].width)

    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "A": (bend90, "W0", "N0"),
        "B": (bend90, "N0", "W0"),
        "-": (wgh, "W0", "E0"),
        "|": (wgv, "W0", "E0"),
    }

    # Generate the sequence of staircases
    s = "-A|B" * n_steps + "-"

    # Create the component from the sequence
    c = component_sequence(s, string_to_device_in_out_ports, start_orientation=0)
    c.update_settings(n_bends=2 * n_steps)
    return c


@pp.autoname
def cutback_bend180(
    bend180=pp.c.bend_euler180,
    straight_length=5.0,
    n_steps=6,
    cols=6,
    spacing=3,
    waveguide_factory=waveguide,
):
    """
      _
    _| |_|

    this is a stair

    _: this is a step


    """
    bend180 = pp.call_if_func(bend180)

    wg = waveguide_factory(length=straight_length, width=bend180.ports["W0"].width)
    wg_vertical = waveguide_factory(
        length=2 * bend180.size_info.width + straight_length + spacing,
        width=bend180.ports["W0"].width,
    )

    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "D": (bend180, "W0", "W1"),
        "C": (bend180, "W1", "W0"),
        "-": (wg, "W0", "E0"),
        "|": (wg_vertical, "W0", "E0"),
    }

    # Generate the sequence of staircases
    s = ""
    for i in range(cols):
        if i % 2 == 0:  # even row
            s += "D-C-" * n_steps + "|"
        else:
            s += "C-D-" * n_steps + "|"
    s = s[:-1]

    # Create the component from the sequence
    c = component_sequence(s, string_to_device_in_out_ports, start_orientation=0)
    c.update_settings(n_bends=n_steps * cols * 2 + cols * 2 - 2)
    return c


@pp.autoname
def cutback_bend_circular(bend_radius=10.0, n_steps=3, n_stairs=4):
    bend90 = bend_circular(radius=bend_radius)
    c = cutback_bend(bend90=bend90, n_steps=n_steps, n_stairs=n_stairs)
    cc = add_fiber_array(c, optical_routing_type=1)
    return cc


if __name__ == "__main__":
    # c = cutback_bend_circular(n_steps=7, n_stairs=4, bend_radius=5) #62
    # c = cutback_bend_circular(n_steps=14, n_stairs=4) #118
    # c = cutback_bend_circular(n_steps=3, n_stairs=4)  # 30
    c = cutback_bend180()
    # c = cutback_bend90()
    pp.show(c)



from typing import List
from omegaconf.listconfig import ListConfig
from phidl.device_layout import Label
from pp.add_labels import get_input_label
from pp.component import ComponentReference
from pp.port import Port


def get_input_labels(
    io_gratings: List[ComponentReference],
    ordered_ports: List[Port],
    component_name: str,
    layer_label: ListConfig,
    gc_port_name: str,
) -> List[Label]:
    elements = []
    for i, g in enumerate(io_gratings):
        label = get_input_label(
            port=ordered_ports[i],
            gc=g,
            gc_index=i,
            component_name=component_name,
            layer_label=layer_label,
            gc_port_name=gc_port_name,
        )
        elements += [label]

    return elements



from numpy import pi, sin, sqrt

neff_ridge = 2.8
neff_shallow = 2.5


def calc_curved_grating_periods(
    fiber_angle=15, wavelength=1.55, n_slab=(neff_ridge + neff_shallow) / 2, n_clad=1.0
):
    """
    The following function calculates the confocal grating periods
    n_slab is the "average slab index" of the grating.
    For 220nm silicon it is 2.8, for 150nm it is 2.5.
    The average is approximately 2.65.
    n_clad is the cladding index in which the fiber is located,
    not the index of the layer above the waveguide.
    If the fiber is in air, then it is 1.0.
    If you use an index matching fluid or glue, then it should be 1.45
    """

    DEG2RAD = pi / 180
    cos_fib_angle = sin(DEG2RAD * fiber_angle)
    n2_reduced = n_slab ** 2 - n_clad ** 2 * cos_fib_angle ** 2
    sqrt_n2_reduced = sqrt(n2_reduced)
    h_period = wavelength * (n_slab + n_clad * cos_fib_angle) / n2_reduced
    v_period = wavelength / sqrt_n2_reduced
    return (h_period, v_period)


def calc_grating_period(
    fiber_angle=13.45,
    wavelength=1.55,
    neff_high=neff_ridge,
    neff_low=neff_shallow,
    n_clad=1.45,
):
    # from lumerical slides
    DEG2RAD = pi / 180
    neff = (neff_high + neff_low) / 2
    return wavelength / (neff - sin(DEG2RAD * fiber_angle) * n_clad)


if __name__ == "__main__":
    p = calc_curved_grating_periods()
    p = calc_grating_period()
    print(p)



from typing import List, Union
import numpy as np
from numpy import float64, ndarray, cos, sin
from pp.coord2 import Coord2


RAD2DEG = 180.0 / np.pi
DEG2RAD = 1 / RAD2DEG


def sign_shape(pts):
    pts2 = np.roll(pts, 1, axis=0)
    dx = pts2[:, 0] - pts[:, 0]
    y = pts2[:, 1] + pts[:, 1]
    return np.sign((dx * y).sum())


def area(pts):
    pts2 = np.roll(pts, 1, axis=0)
    dx = pts2[:, 0] - pts[:, 0]
    y = pts2[:, 1] + pts[:, 1]
    return (dx * y).sum() / 2


def manhattan_direction(p0, p1, tol=1e-5):
    """
    """
    dp = p1 - p0
    dx, dy = dp[0], dp[1]
    if abs(dx) < tol:
        sx = 0
    elif dx > 0:
        sx = 1
    else:
        sx = -1

    if abs(dy) < tol:
        sy = 0
    elif dy > 0:
        sy = 1
    else:
        sy = -1
    return np.array((sx, sy))


def remove_flat_angles(points):
    a = angles_deg(np.vstack(points))
    da = a - np.roll(a, 1)
    da = np.mod(np.round(da, 3), 180)

    # To make sure we do not remove points at the edges
    da[0] = 1
    da[-1] = 1

    to_rm = list(np.where(np.abs(da[:-1]) < 1e-9)[0])
    if type(points) == list:
        while to_rm:
            i = to_rm.pop()
            points.pop(i)

    else:
        points = points[da != 0]

    return points


def remove_identicals(pts, grids_per_unit=1000, closed=True):
    if len(pts) > 1:
        identicals = np.prod(abs(pts - np.roll(pts, -1, 0)) < 0.5 / grids_per_unit, 1)
        if not closed:
            identicals[-1] = False
        pts = np.delete(pts, identicals.nonzero()[0], 0)
    return pts


def centered_diff(a: ndarray) -> ndarray:
    d = (np.roll(a, -1, axis=0) - np.roll(a, 1, axis=0)) / 2
    return d[1:-1]


def centered_diff2(a: ndarray) -> ndarray:
    d = (np.roll(a, -1, axis=0) - a) - (a - np.roll(a, 1, axis=0))
    return d[1:-1]


def curvature(points: ndarray, t: ndarray) -> ndarray:
    """

    Args:
        points : numpy.array shape (n, 2)
        t: numpy.array of size n

    Return:
        The curvature at each point

    Computes the curvature at every point excluding the first and last point

    For a planar curve parametrized as P(t) = (x(t), y(t)), the curvature is given
    by (x' y'' - x'' y' ) / (x' **2 + y' **2)**(3/2)

    """

    # Use centered difference for derivative
    dt = centered_diff(t)
    dp = centered_diff(points)
    dp2 = centered_diff2(points)

    dx = dp[:, 0] / dt
    dy = dp[:, 1] / dt

    dx2 = dp2[:, 0] / dt ** 2
    dy2 = dp2[:, 1] / dt ** 2

    curv = (dx * dy2 - dx2 * dy) / (dx ** 2 + dy ** 2) ** (3 / 2)
    return curv


def radius_of_curvature(points, t):
    return 1 / curvature(points, t)


def path_length(points: ndarray) -> float64:
    """
    Args:
        points: <np.array>
            With shape (N, 2) representing N points with coordinates x, y

    Returns:
        <float> The path length
    """

    dpts = points[1:, :] - points[:-1, :]
    _d = dpts ** 2
    return np.sum(np.sqrt(_d[:, 0] + _d[:, 1]))


def snap_angle(a: float64) -> int:
    """
    a: angle in deg
    Return angle snapped along manhattan angle
    """
    a = a % 360
    if -45 < a < 45:
        _a = 0
    elif 45 < a < 135:
        _a = 90
    elif 135 < a < 225:
        _a = 180
    elif 225 < a < 315:
        _a = 270
    else:
        _a = 0
    return _a


def angles_rad(pts: ndarray) -> ndarray:
    """ returns the angles (radians) of the connection between each point and the next """
    _pts = np.roll(pts, -1, 0)
    radians = np.arctan2(_pts[:, 1] - pts[:, 1], _pts[:, 0] - pts[:, 0])
    return radians


def angles_deg(pts: ndarray) -> ndarray:
    """ returns the angles (degrees) of the connection between each point and the next """
    return angles_rad(pts) * RAD2DEG


def extrude_path(
    points: Union[List[Coord2], ndarray],
    width: Union[float64, float],
    with_manhattan_facing_angles: bool = True,
    spike_length: Union[float64, int, float] = 0,
    start_angle: None = None,
    end_angle: None = None,
    grid: float = 0.001,
) -> ndarray:
    """
    Extrude a path of width `width` along a curve defined by `points`

    Args:
        points: numpy 2D array of shape (N, 2)
        width: float

    Return
        numpy 2D array of shape (2*N, 2)
    """

    if type(points) == list:
        points = np.stack([(p[0], p[1]) for p in points], axis=0)

    a = angles_deg(points)
    if with_manhattan_facing_angles:
        _start_angle = snap_angle(a[0] + 180)
        _end_angle = snap_angle(a[-2])
    else:
        _start_angle = a[0] + 180
        _end_angle = a[-2]

    start_angle = start_angle if start_angle is not None else _start_angle
    end_angle = end_angle if end_angle is not None else _end_angle

    a2 = angles_rad(points) * 0.5
    a1 = np.roll(a2, 1)

    a2[-1] = end_angle * DEG2RAD - a2[-2]
    a1[0] = start_angle * DEG2RAD - a1[1]

    a_plus = a2 + a1
    cos_a_min = np.cos(a2 - a1)
    offsets = np.column_stack((-sin(a_plus) / cos_a_min, cos(a_plus) / cos_a_min)) * (
        0.5 * width
    )

    points_back = np.flipud(points - offsets)
    if spike_length != 0:
        d = spike_length
        a_start = start_angle * DEG2RAD
        a_end = end_angle * DEG2RAD
        p_start_spike = points[0] + d * np.array([[cos(a_start), sin(a_start)]])
        p_end_spike = points[-1] + d * np.array([[cos(a_end), sin(a_end)]])

        pts = np.vstack((p_start_spike, points + offsets, p_end_spike, points_back))
    else:
        pts = np.vstack((points + offsets, points_back))

    pts = np.round(pts / grid) * grid

    return pts


def polygon_grow(polygon, offset):
    """
    polygon has to be a closed shape
    """
    s = remove_identicals(polygon)
    s = remove_flat_angles(s)
    s = np.vstack([s, s[0]])
    if len(s) <= 1:
        return s

    # Make sure the shape is oriented in the correct direction for scaling
    ss = sign_shape(s)
    offset = -ss * offset

    a2 = angles_rad(s) * 0.5
    a1 = np.roll(a2, 1)

    a2[-1] = a2[0]
    a1[0] = a1[-1]

    a = a2 + a1
    c_minus = cos(a2 - a1)
    offsets = np.column_stack((-sin(a) / c_minus, cos(a) / c_minus)) * offset
    # compute offsets from each point

    pts = s + offsets
    return pts



""" merges mask metadata with test and data analysis protocols

-------------------------------------
config.yml

test_protocols:
    passive_optical_te_coarse:
        wl_min:
        wl_max:
        wl_step:
        polarization: te

    passive_optical_tm_coarse:
        wl_min:
        wl_max:
        wl_step:
        polarization: tm
    ...

-------------------------------------
does.yml

doe01:
    instances:
        - cell_name1, x1, y1
        - cell_name2, x2, y2
        - cell_name3, x3, y3

    test_protocols:
        - passive_optical_te_coarse

doe02:
    instances:
        - cell_name21, x21, y21
        - cell_name22, x22, y22
        - cell_name23, x23, y23

    test_protocols:
        - passive_optical_te_coarse
    ...
"""

import pathlib
import json
import yaml
from pp.config import CONFIG


def parse_csv_data(csv_labels_path):
    with open(csv_labels_path) as f:
        # Get all lines
        lines = [line.replace("\n", "") for line in f.readlines()]

        # Ignore labels for metrology structures
        lines = [line for line in lines if not line.startswith("METR_")]

        # Split lines in fields
        lines = [line.split(",") for line in lines]

        lines = [[s.strip() for s in splitted if s.strip()] for splitted in lines]

        # Remove empty lines
        lines = [line for line in lines if line]
    return lines


def get_cell_from_label(label):
    """ get cell name from the label (cell_name is in parenthesis)
    it
    """
    cell_name = label.split("(")[1].split(")")[0]
    if cell_name.startswith("loopback"):
        cell_name = "_".join(cell_name.split("_")[1:])
    return cell_name


def load_json(filepath):
    with open(filepath) as f:
        data = json.load(f)
    return data


def load_yaml(filepath):
    with open(filepath) as f:
        data = yaml.safe_load(f)
    return data


def merge_test_metadata(gdspath=CONFIG["mask_gds"], labels_prefix="opt"):
    """ from a gds mask combines test_protocols and labels positions for each DOE
    Do a map cell: does
    Usually each cell will have only one DOE. But in general it should be allowed for a cell to belong to multiple DOEs

    Args:
        gdspath

    Returns:
        saves json file with merged metadata

    """
    gdspath = pathlib.Path(gdspath)
    mask_json_path = gdspath.with_suffix(".json")
    csv_labels_path = gdspath.with_suffix(".csv")
    output_tm_path = gdspath.with_suffix(".tp.json")

    assert mask_json_path.exists(), f"missing mask JSON metadata {mask_json_path}"
    assert csv_labels_path.exists(), f"missing CSV labels {csv_labels_path}"

    metadata = load_json(mask_json_path)
    labels_list = parse_csv_data(csv_labels_path)

    does = metadata.pop("does")
    cells = metadata.pop("cells")

    c = {}

    for label, x, y in labels_list:
        cell = get_cell_from_label(label)
        c[cell] = c.get(cell, dict())
        c[cell][label] = dict(x=x, y=y)

    d = dict(cells_to_test=c, metadata=metadata, does=does, cells=cells)

    with open(output_tm_path, "w") as json_out:
        json.dump(d, json_out, indent=2)

    return metadata


if __name__ == "__main__":
    from pp import CONFIG

    gdspath = CONFIG["repo_path"] / "samples" / "mask" / "build" / "mask" / "mask.gds"
    d = merge_test_metadata(gdspath)
    print(d)



import numpy as np
import pp
from pp.geo_utils import extrude_path
from pp.geo_utils import DEG2RAD
from pp.layers import LAYER
from numpy import float64, ndarray
from pp.component import Component
from typing import Tuple, Union


def ellipse_arc(
    a: float64,
    b: float64,
    x0: float64,
    theta_min: float,
    theta_max: float,
    angle_step: float = 0.5,
) -> ndarray:
    theta = np.arange(theta_min, theta_max + angle_step, angle_step) * DEG2RAD
    xs = a * np.cos(theta) + x0
    ys = b * np.sin(theta)
    return np.column_stack([xs, ys])


def grating_tooth_points(
    ap: float64,
    bp: float64,
    xp: float64,
    width: Union[float64, float],
    taper_angle: float,
    spiked: bool = True,
    angle_step: float = 1.0,
) -> ndarray:
    theta_min = -taper_angle / 2
    theta_max = taper_angle / 2

    backbone_points = ellipse_arc(ap, bp, xp, theta_min, theta_max, angle_step)
    if spiked:
        spike_length = width / 3
    else:
        spike_length = 0.0
    points = extrude_path(
        backbone_points,
        width,
        with_manhattan_facing_angles=False,
        spike_length=spike_length,
    )

    return points


def grating_taper_points(
    a: float64,
    b: float64,
    x0: float64,
    taper_length: float64,
    taper_angle: float,
    wg_width: float,
    angle_step: float = 1.0,
) -> ndarray:
    taper_arc = ellipse_arc(a, b, taper_length, -taper_angle / 2, taper_angle / 2)

    port_position = np.array((x0, 0))
    p0 = port_position + (0, wg_width / 2)
    p1 = port_position + (0, -wg_width / 2)
    points = np.vstack([p0, p1, taper_arc])
    return points


@pp.autoname
def grating_coupler_elliptical_tm(
    taper_length: float = 16.6,
    taper_angle: float = 30.0,
    lambda_c: float = 1.554,
    fiber_angle: float = 15.0,
    grating_line_width: float = 0.707,
    wg_width: float = 0.5,
    neff: float = 1.8,  # tooth effective index
    layer: Tuple[int, int] = LAYER.WG,
    n_periods: int = 16,
    **kwargs
) -> Component:
    """

    Args:
        neff: tooth effective index

    .. plot::
      :include-source:

      import pp

      c = pp.c.grating_coupler_elliptical_tm()
      pp.plotgds(c)

    """
    return grating_coupler_elliptical(
        polarization="tm",
        taper_length=taper_length,
        taper_angle=taper_angle,
        lambda_c=lambda_c,
        fiber_angle=fiber_angle,
        grating_line_width=grating_line_width,
        wg_width=wg_width,
        neff=neff,
        layer=layer,
        n_periods=n_periods,
        big_last_tooth=False,
        **kwargs
    )


@pp.autoname
def grating_coupler_elliptical_te(
    taper_length: float = 16.6,
    taper_angle: float = 40.0,
    lambda_c: float = 1.554,
    fiber_angle: float = 15.0,
    grating_line_width: float = 0.343,
    wg_width: float = 0.5,
    neff: float = 2.638,  # tooth effective index
    layer: Tuple[int, int] = LAYER.WG,
    p_start: int = 26,
    n_periods: int = 24,
    **kwargs
) -> Component:
    return grating_coupler_elliptical(
        polarization="te",
        taper_length=taper_length,
        taper_angle=taper_angle,
        lambda_c=lambda_c,
        fiber_angle=fiber_angle,
        grating_line_width=grating_line_width,
        wg_width=wg_width,
        neff=neff,
        layer=layer,
        p_start=p_start,
        n_periods=n_periods,
        **kwargs
    )


@pp.autoname
def grating_coupler_elliptical(
    polarization: str,
    taper_length: float = 16.6,
    taper_angle: float = 30.0,
    lambda_c: float = 1.554,
    fiber_angle: float = 15.0,
    grating_line_width: float = 0.343,
    wg_width: float = 0.5,
    neff: float = 2.638,  # tooth effective index
    layer: Tuple[int, int] = LAYER.WG,
    p_start: int = 26,
    n_periods: int = 30,
    big_last_tooth: bool = False,
    layer_slab: Tuple[int, int] = LAYER.SLAB150,
    with_fiber_marker: bool = True,
) -> Component:
    """

    Args:
        taper_length: taper length from waveguide I/O
        taper_angle: grating flare angle
        lambda_c: grating transmission central wavelength
        fiber_angle: fibre polish angle in degrees
        grating_line_width
        wg_width: waveguide width
        neff: 2.638  # tooth effective index
        layer: LAYER.WG
        p_start: 26  # first tooth
        n_periods: 16  # number of periods

    .. plot::
      :include-source:

      import pp

      c = pp.c.grating_coupler_elliptical_te()
      pp.plotgds(c)
    """

    # Define some constants
    nc = 1.443  # cladding index

    # Compute some ellipse parameters
    sthc = np.sin(fiber_angle * DEG2RAD)
    d = neff ** 2 - nc ** 2 * sthc ** 2
    a1 = lambda_c * neff / d
    b1 = lambda_c / np.sqrt(d)
    x1 = lambda_c * nc * sthc / d

    a1 = round(a1, 3)
    b1 = round(b1, 3)
    x1 = round(x1, 3)

    period = a1 + x1

    c = pp.Component()
    c.polarization = polarization
    c.wavelength = int(lambda_c * 1e3)

    # Make each grating line
    for p in range(p_start, p_start + n_periods + 1):
        pts = grating_tooth_points(
            p * a1, p * b1, p * x1, grating_line_width, taper_angle
        )
        c.add_polygon(pts, layer)

    # Make the taper
    p_taper = p_start - 1
    p_taper_eff = p_taper
    a_taper = a1 * p_taper_eff
    b_taper = b1 * p_taper_eff
    x_taper = x1 * p_taper_eff

    x_output = a_taper + x_taper - taper_length + grating_line_width / 2
    pts = grating_taper_points(
        a_taper, b_taper, x_output, x_taper, taper_angle, wg_width=wg_width
    )
    c.add_polygon(pts, layer)

    # Superimpose a tooth without spikes at end of taper to match the period.
    pts = grating_tooth_points(
        a_taper, b_taper, x_taper, grating_line_width, taper_angle, spiked=False
    )
    c.add_polygon(pts, layer)

    # Add last "large tooth" after the standard grating teeth
    w = 1.0
    L = (
        period * (p_start + n_periods)
        + grating_line_width / 2
        + period
        - grating_line_width
        + w / 2
    )

    if big_last_tooth:
        a = L / (1 + x1 / a1)
        b = b1 / a1 * a
        x = x1 / a1 * a

        pts = grating_tooth_points(a, b, x, w, taper_angle, spiked=False)
        c.add_polygon(pts, layer)

    # Move waveguide I/O to (0, 0)
    c.move((-x_output, 0))

    if polarization.lower() == "te":
        polarization_marker_layer = pp.LAYER.polarization_te
    else:
        polarization_marker_layer = pp.LAYER.polarization_tm

    if with_fiber_marker:
        circle = pp.c.circle(radius=17 / 2, layer=polarization_marker_layer)
        circle_ref = c.add_ref(circle)
        circle_ref.movex(taper_length + period * n_periods / 2)

    # Add port
    c.add_port(name="W0", midpoint=[0, 0], width=wg_width, orientation=180, layer=layer)

    # Add shallow etch
    _rl = L + grating_line_width + 2.0
    _rhw = _rl * np.tan(fiber_angle * DEG2RAD) + 2.0

    if layer_slab:
        c.add_polygon([(0, _rhw), (_rl, _rhw), (_rl, -_rhw), (0, -_rhw)], LAYER.SLAB150)

    return c


if __name__ == "__main__":
    c = grating_coupler_elliptical_tm()
    c = grating_coupler_elliptical_te(layer_slab=None, with_fiber_marker=False)
    print(c.polarization)
    pp.write_gds(c)
    pp.show(c)



from pp.components import bend_circular
from pp.components import waveguide
from pp.routing.manhattan import round_corners


def gen_loopback(
    start_port,
    end_port,
    gc,
    grating_separation=127.0,
    gc_rotation=-90,
    gc_port_name="W0",
    bend_radius_align_ports=10.0,
    bend_factory=bend_circular,
    waveguide_factory=waveguide,
    y_bot_align_route=None,
):
    """
    Add a loopback (grating coupler align reference) to a start port and and end port
    Input grating generated on the left of start_port
    Output grating generated on the right of end_port

    .. code::
        __________________________________________
        | separation  |            |              |
        |             |            |              |
       GC          start_port  end_port          GC
    """

    gc = gc() if callable(gc) else gc

    if hasattr(start_port, "y"):
        y0 = start_port.y
    else:
        y0 = start_port[1]

    if hasattr(start_port, "x"):
        x0 = start_port.x - grating_separation
    else:
        x0 = start_port[0] - grating_separation

    if hasattr(end_port, "x"):
        x1 = end_port.x + grating_separation
    else:
        x1 = end_port[0] + grating_separation

    gca1, gca2 = [
        gc.ref(position=(x, y0), rotation=gc_rotation, port_id=gc_port_name)
        for x in [x0, x1]
    ]

    gsi = gc.size_info
    p0 = gca1.ports[gc_port_name].position
    p1 = gca2.ports[gc_port_name].position
    a = bend_radius_align_ports + 0.5
    b = max(2 * a, grating_separation / 2)
    y_bot_align_route = (
        y_bot_align_route if y_bot_align_route is not None else -gsi.width - 5.0
    )

    route = [
        p0,
        p0 + (0, a),
        p0 + (b, a),
        p0 + (b, y_bot_align_route),
        p1 + (-b, y_bot_align_route),
        p1 + (-b, a),
        p1 + (0, a),
        p1,
    ]
    bend90 = bend_factory(radius=bend_radius_align_ports)
    loop_back = round_corners(route, bend90, waveguide_factory)
    elements = [gca1, gca2, loop_back]
    return elements


if __name__ == "__main__":
    import pp

    c = waveguide()
    c.add(gen_loopback(c.ports["W0"], c.ports["E0"], gc=pp.c.grating_coupler_te))
    pp.show(c)



import os
from glob import glob

from pp.config import logging, CONFIG


def merge_markdown(
    reports_directory=CONFIG["doe_directory"],
    mdpath=CONFIG["mask_directory"] / "report.md",
):
    """ Merges all individual markdown reports (.md) into a single markdown
    you can add a report:[Capacitors, Diodes...] in config.yml to define the merge order
    """
    logging.debug("Merging Markdown files:")

    with open(mdpath, "w") as f:

        def wl(line="", eol="\n"):
            f.write(line + eol)

        reports = sorted(glob(os.path.join(reports_directory, "*.md")))
        for filename in reports:
            with open(filename) as infile:
                for line in infile:
                    f.write(line)

    logging.info(f"Wrote {mdpath}")


if __name__ == "__main__":
    reports_directory = CONFIG["samples_path"] / "mask" / "does"
    merge_markdown(reports_directory)



from typing import Callable, List, Tuple
from pp.component import Component
from pp.name import autoname
from pp.layers import LAYER
from pp.drc import assert_on_1nm_grid
from pp.components.coupler_symmetric import coupler_symmetric
from pp.components.coupler_straight import coupler_straight


@autoname
def coupler(
    wg_width: float = 0.5,
    gap: float = 0.236,
    length: float = 20.007,
    coupler_symmetric_factory: Callable = coupler_symmetric,
    coupler_straight: Callable = coupler_straight,
    layer: Tuple[int, int] = LAYER.WG,
    layers_cladding: List[Tuple[int, int]] = [LAYER.WGCLAD],
    cladding_offset: int = 3,
) -> Component:
    r""" symmetric coupler

    Args:
        gap
        length
        coupler_symmetric_factory
        coupler_straight

    .. code::

       W1 __                           __ E1
            \                         /
             \        length         /
              ======================= gap
             /                        \
           _/                          \_
        W0                              E0

            coupler_straight  coupler_symmetric_factory

    .. plot::
      :include-source:

      import pp

      c = pp.c.coupler(gap=0.2, length=10)
      pp.plotgds(c)

    """
    assert_on_1nm_grid(length)
    assert_on_1nm_grid(gap)
    c = Component()

    sbend = coupler_symmetric_factory(
        gap=gap,
        wg_width=wg_width,
        layer=layer,
        layers_cladding=layers_cladding,
        cladding_offset=cladding_offset,
    )

    sr = c << sbend
    sl = c << sbend
    cs = c << coupler_straight(
        length=length,
        gap=gap,
        width=wg_width,
        layer=layer,
        layers_cladding=layers_cladding,
        cladding_offset=cladding_offset,
    )
    sl.connect("W0", destination=cs.ports["W0"])
    sr.connect("W0", destination=cs.ports["E0"])

    c.add_port("W1", port=sl.ports["E0"])
    c.add_port("W0", port=sl.ports["E1"])
    c.add_port("E0", port=sr.ports["E0"])
    c.add_port("E1", port=sr.ports["E1"])

    c.absorb(sl)
    c.absorb(sr)
    c.absorb(cs)
    return c


if __name__ == "__main__":
    import pp

    c = coupler(length=10, pins=True)
    print(c.settings_changed)
    pp.show(c)



import numpy as np
import pp
from pp.routing.connect import connect_strip
from pp.routing.connect import connect_elec
from pp.port import is_electrical_port
from pp.port import flipped

BEND_RADIUS = pp.config.BEND_RADIUS


def route_elec_ports_to_side(ports, side="north", wire_sep=20.0, x=None, y=None):
    return route_ports_to_side(
        ports, side=side, bend_radius=0, separation=wire_sep, x=x, y=y
    )


def route_ports_to_side(
    ports, side="north", x=None, y=None, routing_func=None, **kwargs
):
    """ Routes ports to a given side

    Args:
        ports: the list of ports to be connected to the side
            can also be a dictionnary, a <pp.Component> or a phidl
            <ComponentReference>
        side should be 'north', 'south', 'east' or 'west'

        x: only for east/west side routing: the x position where the ports should be sent
            If None, will use the eastest/westest value

        y: only for south/north side routing: the y position where the ports should be send
            If None, will use the southest/northest value

        routing_func: the routing function. By default uses either `connect_elec`
        or `connect_strip` depending on the ports layer.

        kwargs: may include:
            `bend_radius`
            `extend_bottom`, `extend_top` for east/west routing
            `extend_left`, `extend_right` for south/north routing
    """

    if not ports:
        return [], []

    # Accept list of ports, Component or dict of ports
    if isinstance(ports, dict):
        ports = list(ports.values())

    elif isinstance(ports, pp.Component) or isinstance(ports, pp.ComponentReference):
        ports = list(ports.ports.values())

    # Convenient default selection for connection function point to point
    if routing_func is None:
        if is_electrical_port(ports[0]):
            routing_func = connect_elec
        else:
            routing_func = connect_strip

    # Choose which
    if side in ["north", "south"]:
        func_route = connect_ports_to_y
        if y is not None:
            xy = y
        else:
            xy = side

    elif side in ["west", "east"]:
        if x is not None:
            xy = x
        else:
            xy = side

        func_route = connect_ports_to_x

    return func_route(ports, xy, routing_func=routing_func, **kwargs)


def route_ports_to_north(list_ports, **kwargs):
    return route_ports_to_side(list_ports, side="north", **kwargs)


def route_ports_to_south(list_ports, **kwargs):
    return route_ports_to_side(list_ports, side="south", **kwargs)


def route_ports_to_west(list_ports, **kwargs):
    return route_ports_to_side(list_ports, side="west", **kwargs)


def route_ports_to_east(list_ports, **kwargs):
    return route_ports_to_side(list_ports, side="east", **kwargs)


def connect_ports_to_x(
    list_ports,
    x="east",
    separation=10.0,
    bend_radius=BEND_RADIUS,
    extend_bottom=0,
    extend_top=0,
    extension_length=0,
    y0_bottom=None,
    y0_top=None,
    routing_func=connect_strip,
    routing_func_args={},
    backward_port_side_split_index=0,
):
    """
     * ``list_ports``: reasonably well behaved list of ports
            i.e
            ports facing north ports are norther than any other ports
            ports facing south ports are souther ...
            ports facing west ports are the wester ...
            ports facing east ports are the easter ...

     * ``x``: float or string:
                if float: x coordinate to which the ports will be routed
                if string: "east" -> route to east
                if string: "west" -> route to west

     * ``backward_port_side_split_index``: integer
            this integer represents and index in the list of backwards ports
                (bottom to top)
            all ports with an index strictly lower or equal are routed bottom
            all ports with an index larger or equal are routed top

    Returns:
        - a list of connectors which can be added to an element list
        - a list of the new optical ports

    """

    north_ports = [p for p in list_ports if p.angle == 90]
    south_ports = [p for p in list_ports if p.angle == 270]
    east_ports = [p for p in list_ports if p.angle == 0]
    west_ports = [p for p in list_ports if p.angle == 180]

    epsilon = 1.0
    a = epsilon + max(bend_radius, separation)
    xs = [p.x for p in list_ports]
    ys = [p.y for p in list_ports]

    if y0_bottom is None:
        y0_bottom = min(ys) - a
    y0_bottom -= extend_bottom

    if y0_top is None:
        y0_top = max(ys) + a
    y0_top += extend_top

    if x == "west" and extension_length > 0:
        extension_length = -extension_length

    if x == "east":
        x = max([p.x for p in list_ports]) + a
    elif x == "west":
        x = min([p.x for p in list_ports]) - a
    elif type(x) == float:
        pass
    else:
        pass
        # raise ValueError('``x`` should be a float or "east" or "west"')

    sort_key_west_to_east = lambda p: p.x
    sort_key_east_to_west = lambda p: -p.x

    sort_key_south_to_north = lambda p: p.y
    sort_key_north_to_south = lambda p: -p.y

    if x < min(xs):
        sort_key_north = sort_key_west_to_east
        sort_key_south = sort_key_west_to_east
        forward_ports = west_ports
        backward_ports = east_ports
        angle = 0

    elif x > max(xs):
        sort_key_south = sort_key_east_to_west
        sort_key_north = sort_key_east_to_west
        forward_ports = east_ports
        backward_ports = west_ports
        angle = 180
    else:
        raise ValueError("x should be either to the east or to the west of all ports")
    """
    First route the bottom-half of the back ports
        (back ports are the one facing opposite side of x)
    Then route the south ports
    then the front ports
    then the north ports
    """

    # forward_ports.sort()
    north_ports.sort(key=sort_key_north)
    south_ports.sort(key=sort_key_south)
    forward_ports.sort(key=sort_key_south_to_north)

    backward_ports.sort(key=sort_key_south_to_north)
    backward_ports_thru_south = backward_ports[0:backward_port_side_split_index]
    backward_ports_thru_north = backward_ports[backward_port_side_split_index:]
    backward_ports_thru_south.sort(key=sort_key_south_to_north)
    backward_ports_thru_north.sort(key=sort_key_north_to_south)

    elements = []
    ports = []

    def add_port(p, y, l_elements, l_ports, start_straight=0.01):
        new_port = p._copy()
        new_port.angle = angle
        new_port.position = (x + extension_length, y)
        l_elements += [
            routing_func(
                p,
                new_port,
                start_straight=start_straight,
                bend_radius=bend_radius,
                **routing_func_args
            )
        ]
        l_ports += [flipped(new_port)]

    y_optical_bot = y0_bottom
    for p in south_ports:
        add_port(p, y_optical_bot, elements, ports)
        y_optical_bot -= separation

    for p in forward_ports:
        add_port(p, p.y, elements, ports)

    y_optical_top = y0_top
    for p in north_ports:
        add_port(p, y_optical_top, elements, ports)
        y_optical_top += separation

    start_straight = 0.01
    start_straight0 = 0
    max_x = max(xs)
    min_x = min(xs)

    for p in backward_ports_thru_north:
        # Extend ports if necessary
        if angle == 0 and p.x < max_x:
            start_straight0 = max_x - p.x
        elif angle == 180 and p.x > min_x:
            start_straight0 = p.x - min_x
        else:
            start_straight0 = 0

        add_port(
            p,
            y_optical_top,
            elements,
            ports,
            start_straight=start_straight + start_straight0,
        )
        y_optical_top += separation
        start_straight += separation

    start_straight = 0.01
    start_straight0 = 0
    for p in backward_ports_thru_south:
        # Extend ports if necessary
        if angle == 0 and p.x < max_x:
            start_straight0 = max_x - p.x
        elif angle == 180 and p.x > min_x:
            start_straight0 = p.x - min_x
        else:
            start_straight0 = 0

        add_port(
            p,
            y_optical_bot,
            elements,
            ports,
            start_straight=start_straight + start_straight0,
        )
        y_optical_bot -= separation
        start_straight += separation

    return elements, ports


def connect_ports_to_y(
    list_ports,
    y="north",
    separation=10.0,
    bend_radius=BEND_RADIUS,
    x0_left=None,
    x0_right=None,
    extension_length=0,
    extend_left=0,
    extend_right=0,
    routing_func=connect_strip,
    routing_func_args={},
    backward_port_side_split_index=0,
):
    """
     * ``list_ports``: reasonably well behaved list of ports
            i.e
            ports facing north ports are norther than any other ports
            ports facing south ports are souther ...
            ports facing west ports are the wester ...
            ports facing east ports are the easter ...

     * ``y``: float or string:
                if float: y coordinate to which the ports will be routed
                if string: "north" -> route to north
                if string: "south" -> route to south

     * ``backward_port_side_split_index``: integer
            this integer represents and index in the list of backwards ports
                (sorted from left to right)
            all ports with an index strictly larger are routed right
            all ports with an index lower or equal are routed left

    Returns:
        - a list of connectors which can be added to an element list
        - a list of the new optical ports

    """

    if y == "south" and extension_length > 0:
        extension_length = -extension_length

    da = 45
    north_ports = [p for p in list_ports if p.angle > 90 - da and p.angle < 90 + da]
    south_ports = [p for p in list_ports if p.angle > 270 - da and p.angle < 270 + da]
    east_ports = [p for p in list_ports if p.angle < da or p.angle > 360 - da]
    west_ports = [p for p in list_ports if p.angle < 180 + da and p.angle > 180 - da]

    epsilon = 1.0
    a = bend_radius + max(bend_radius, separation)
    xs = [p.x for p in list_ports]
    ys = [p.y for p in list_ports]

    x0_right = x0_right or max(xs) + a
    x0_right += extend_right
    x0_left = x0_left or min(xs) - a
    x0_left -= extend_left

    if y == "north":
        y = (
            max([p.y + a * np.abs(np.cos(p.angle * np.pi / 180)) for p in list_ports])
            + epsilon
        )
    elif y == "south":
        y = (
            min([p.y - a * np.abs(np.cos(p.angle * np.pi / 180)) for p in list_ports])
            - epsilon
        )
    elif type(y) == float:
        pass
    else:
        pass
        # raise ValueError('``y`` should be a float or "north" or "south"')

    sort_key_west_to_east = lambda p: p.x
    sort_key_east_to_west = lambda p: -p.x
    sort_key_south_to_north = lambda p: p.y
    sort_key_north_to_south = lambda p: -p.y

    if y <= min(ys):
        sort_key_east = sort_key_south_to_north
        sort_key_west = sort_key_south_to_north
        forward_ports = south_ports
        backward_ports = north_ports
        angle = 90.0

    elif y >= max(ys):
        sort_key_west = sort_key_north_to_south
        sort_key_east = sort_key_north_to_south
        forward_ports = north_ports
        backward_ports = south_ports
        angle = -90.0
    else:
        raise ValueError("y should be either to the north or to the south of all ports")
    """
    First route the bottom-half of the back ports
        (back ports are the one facing opposite side of x)
    Then route the south ports
    then the front ports
    then the north ports
    """

    west_ports.sort(key=sort_key_west)
    east_ports.sort(key=sort_key_east)
    forward_ports.sort(key=sort_key_west_to_east)
    backward_ports.sort(key=sort_key_east_to_west)

    backward_ports.sort(key=sort_key_west_to_east)
    backward_ports_thru_west = backward_ports[0:backward_port_side_split_index]
    backward_ports_thru_east = backward_ports[backward_port_side_split_index:]

    backward_ports_thru_west.sort(key=sort_key_west_to_east)
    backward_ports_thru_east.sort(key=sort_key_east_to_west)

    elements = []
    ports = []

    def add_port(p, x, l_elements, l_ports, start_straight=0.01):
        new_port = p._copy()
        new_port.angle = angle
        new_port.position = (x, y + extension_length)

        if np.sum(np.abs((new_port.position - p.position) ** 2)) < 1e-12:
            l_ports += [flipped(new_port)]
            return

        try:
            l_elements += [
                routing_func(
                    p,
                    new_port,
                    start_straight=start_straight,
                    bend_radius=bend_radius,
                    **routing_func_args
                )
            ]
            l_ports += [flipped(new_port)]

        except Exception as e:
            print("**************************")
            print("Could not connect")
            print(p)
            print(new_port)
            print("**************************")
            raise e

    x_optical_left = x0_left
    for p in west_ports:
        add_port(p, x_optical_left, elements, ports)
        x_optical_left -= separation

    for p in forward_ports:
        add_port(p, p.x, elements, ports)

    x_optical_right = x0_right
    for p in east_ports:
        add_port(p, x_optical_right, elements, ports)
        x_optical_right += separation

    start_straight = 0.01
    for p in backward_ports_thru_east:
        add_port(p, x_optical_right, elements, ports, start_straight=start_straight)
        x_optical_right += separation
        start_straight += separation

    start_straight = 0.01
    for p in backward_ports_thru_west:
        add_port(p, x_optical_left, elements, ports, start_straight=start_straight)
        x_optical_left -= separation
        start_straight += separation

    return elements, ports


def demo():
    from pp.component import Component
    from pp.layers import LAYER

    def dummy():
        cmp = Component()
        xs = [0.0, 10.0, 25.0, 50.0]
        ys = [0.0, 10.0, 25.0, 50.0]
        a = 5
        xl = min(xs) - a
        xr = max(xs) + a
        yb = min(ys) - a
        yt = max(ys) + a

        cmp.add_polygon([(xl, yb), (xl, yt), (xr, yt), (xr, yb)], LAYER.WG)

        for i, y in enumerate(ys):
            p0 = (xl, y)
            p1 = (xr, y)
            cmp.add_port(name="W{}".format(i), midpoint=p0, orientation=180, width=0.5)
            cmp.add_port(name="E{}".format(i), midpoint=p1, orientation=0, width=0.5)

        for i, x in enumerate(xs):
            p0 = (x, yb)
            p1 = (x, yt)
            cmp.add_port(name="S{}".format(i), midpoint=p0, orientation=270, width=0.5)
            cmp.add_port(name="N{}".format(i), midpoint=p1, orientation=90, width=0.5)

        return cmp

    def top_level():
        cmp = Component()
        _dummy_t = dummy()
        sides = ["north", "south", "east", "west"]
        positions = [(0, 0), (400, 0), (400, 400), (0, 400)]
        for pos, side in zip(positions, sides):
            dummy_ref = _dummy_t.ref(position=pos)
            cmp.add(dummy_ref)
            conns, ports = route_ports_to_side(dummy_ref, side)
            for e in conns:
                cmp.add(e)
            for i, p in enumerate(ports):
                cmp.add_port(name="{}{}".format(side[0], i), port=p)
        return cmp

    pp.show(top_level())


if __name__ == "__main__":
    demo()



import gdspy
from pp.compare_cells import hash_cells
from pp.components.mzi2x2 import mzi2x2
import pp


def debug():
    c = mzi2x2()
    h0 = c.hash_geometry()

    gdspath1 = "{}.gds".format(c.name)
    gdspath2 = "{}_2.gds".format(c.name)
    gdspath3 = "{}_3.gds".format(c.name)
    pp.write_gds(c, gdspath1)

    c1 = pp.import_gds(gdspath1, overwrite_cache=True)

    c2 = pp.import_gds(gdspath2, overwrite_cache=True)

    c3 = pp.import_gds(gdspath3, overwrite_cache=True)

    dbg = False
    dh1 = hash_cells(c1, {}, dbg=dbg)
    dh2 = hash_cells(c2, {}, dbg=dbg)
    dh3 = hash_cells(c3, {}, dbg=dbg)

    h1 = dh1[c1.name]
    h2 = dh2[c2.name]
    h3 = dh3[c3.name]
    print(h1)
    print(h2)
    print(h3)

    print(h0)
    print(gdspy.gdsii_hash(gdspath1))
    print(gdspy.gdsii_hash(gdspath2))
    print(gdspy.gdsii_hash(gdspath3))


def test_hash():
    c1 = pp.c.waveguide(length=10)
    c2 = pp.c.waveguide(length=11)
    h1 = c1.hash_geometry()
    h2 = c2.hash_geometry()
    assert h1 != h2


if __name__ == "__main__":
    debug()



"""
In this example we change the layer of the waveguide
"""

import pp


@pp.autoname
def wg(layer=(2, 0), **kwargs):
    return pp.c.waveguide(layer=layer, **kwargs)


if __name__ == "__main__":
    c = wg()
    pp.show(c)



import re
import numpy as np
import pp


def load(
    component=None, filepath=None, dirpath=pp.CONFIG["sp"], numports=None, height_nm=220
):
    """ Load Sparameters from interconnect export

    Args:
        component: instance
        filepath: Sparameters filepath (interconnect format)
        dirpath: path where to look for the Sparameters
        height_nm: height (nm)

    Returns [port_names, F, S]
        port_names: list of strings
        F: frequency 1d np.array
        S: Sparameters np.ndarray matrix

    inspired in https://github.com/BYUCamachoLab/simphony
    the Sparameters file have Lumerical format
    https://support.lumerical.com/hc/en-us/articles/360036107914-Optical-N-Port-S-Parameter-SPAR-INTERCONNECT-Element#toc_5
    """
    if filepath is None:
        assert isinstance(component, pp.Component)
        filepath = component.get_sparameters_path(dirpath=dirpath, height_nm=height_nm)
        numports = len(component.ports)
    assert filepath.exists(), f"Sparameters for {component} not found in {filepath}"
    assert numports > 1, f"{numports} needs to be > 1"

    F = []
    S = []
    port_names = []

    with open(filepath, "r") as fid:
        for i in range(numports):
            port_line = fid.readline()
            m = re.search('\[".*",', port_line)
            if m:
                port = m.group(0)
                port_names.append(port[2:-2])
        line = fid.readline()
        line = fid.readline()
        numrows = int(tuple(line[1:-2].split(","))[0])
        S = np.zeros((numrows, numports, numports), dtype="complex128")
        r = m = n = 0
        for line in fid:
            if line[0] == "(":
                continue
            data = line.split()
            data = list(map(float, data))
            if m == 0 and n == 0:
                F.append(data[0])
            S[r, m, n] = data[1] * np.exp(1j * data[2])
            r += 1
            if r == numrows:
                r = 0
                m += 1
                if m == numports:
                    m = 0
                    n += 1
                    if n == numports:
                        break
    return (port_names, F, S)


if __name__ == "__main__":
    s = load(pp.c.mmi2x2())
    print(s[0])
    # print(s)



import numpy as np
from pp.container import container

import pp
from numpy import float64, ndarray
from typing import List, Optional

DEG2RAD = np.pi / 180


def line(
    p_start: ndarray, p_end: ndarray, width: Optional[float] = None
) -> List[ndarray]:
    if isinstance(p_start, pp.Port):
        width = p_start.width
        p_start = p_start.midpoint

    if isinstance(p_end, pp.Port):
        p_end = p_end.midpoint

    w = width
    angle = np.arctan2(p_end[1] - p_start[1], p_end[0] - p_start[0])
    a = np.pi / 2
    p0 = move_polar_rad_copy(p_start, angle + a, w / 2)
    p1 = move_polar_rad_copy(p_start, angle - a, w / 2)
    p2 = move_polar_rad_copy(p_end, angle - a, w / 2)
    p3 = move_polar_rad_copy(p_end, angle + a, w / 2)
    return [p0, p1, p2, p3]


def move_polar_rad_copy(pos: ndarray, angle: float64, length: float) -> ndarray:
    c = np.cos(angle)
    s = np.sin(angle)
    return pos + length * np.array([c, s])


@pp.autoname
def extend_port(port, length):
    """ returns a port extended by length """
    c = pp.Component()

    # Generate a port extension
    p_start = port.midpoint
    angle = port.angle
    p_end = move_polar_rad_copy(p_start, angle * DEG2RAD, length)
    w = port.width

    _line = line(p_start, p_end, w)

    c.add_polygon(_line, layer=port.layer)
    c.add_port(name="original", port=port)
    c.add_port(
        name=port.name,
        midpoint=p_end,
        width=port.width,
        orientation=port.orientation,
        layer=port.layer,
    )

    return c


@container
def extend_ports(
    component,
    port_list=None,
    length=5,
    extension_factory=None,
    input_port_ext=None,
    output_port_ext=None,
):
    """ returns a component with extended ports
    """
    c = pp.Component(name=component.name + "_e")
    c << component

    port_list = port_list or list(component.ports.keys())

    if extension_factory is None:
        dummy_port = component.ports[port_list[0]]

        def _ext_factory(length, width):
            return pp.c.hline(length=length, width=width, layer=dummy_port.layer)

        extension_factory = _ext_factory

    dummy_ext = extension_factory(length=length, width=0.5)
    port_labels = list(dummy_ext.ports.keys())

    input_port_ext = input_port_ext or port_labels[0]
    output_port_ext = output_port_ext or port_labels[-1]

    for port_label in port_list:
        port = component.ports.get(port_label)
        extension = c << extension_factory(length=length, width=port.width)
        extension.connect(input_port_ext, port)
        c.add_port(port_label, port=extension.ports[output_port_ext])
    return c


def test_extend_ports():
    import pp.components as pc

    c = pc.waveguide()
    ce = extend_ports(c)
    assert len(c.ports) == len(ce.ports)


if __name__ == "__main__":
    test_extend_ports()
    import pp.components as pc

    # c = pc.bend_circular()
    # ce = extend_ports(c, port_list=['W0'])

    c = pc.waveguide(layer=(3, 0))
    ce = extend_ports(c)
    print(ce)
    print(len(ce.ports))
    pp.show(ce)



import pp
from pp import components as pc
from pp.component import Component
from typing import List


@pp.autoname
def litho_calipers(
    notch_size: List[int] = [2, 5],
    notch_spacing: int = 2,
    num_notches: int = 11,
    offset_per_notch: float = 0.1,
    row_spacing: int = 0,
    layer_top: int = 1,
    layer_bottom: int = 2,
) -> Component:
    """ vernier caliper structure to test lithography alignment
    Only the middle finger is aligned and the rest are offset.

    Args:
        notch_size: [xwidth, yheight]
        notch_spacing: 2
        num_notches: 11
        offset_per_notch: 0.1
        row_spacing: 0
        layer_top: 1
        layer_bottom:2

    .. plot::
      :include-source:

      import pp

      c = pp.c.litho_calipers()
      pp.plotgds(c)
    """

    D = pp.Component()
    num_notches_total = num_notches * 2 + 1
    centre_notch = num_notches
    R1 = pc.rectangle(size=(notch_size), layer=layer_top)
    R2 = pc.rectangle(size=(notch_size), layer=layer_bottom)
    for i in range(num_notches_total):
        if i == centre_notch:
            D.add_ref(R1).movex(i * (notch_size[0] + notch_spacing)).movey(
                notch_size[1]
            )
            D.add_ref(R2).movex(
                i * (notch_size[0] + notch_spacing)
                + offset_per_notch * (centre_notch - i)
            ).movey(-2 * notch_size[1] - row_spacing)
        D.add_ref(R1).movex(i * (notch_size[0] + notch_spacing))
        D.add_ref(R2).movex(
            i * (notch_size[0] + notch_spacing) + offset_per_notch * (centre_notch - i)
        ).movey(-notch_size[1] - row_spacing)

    return D


if __name__ == "__main__":
    c = litho_calipers()
    pp.show(c)



import pytest

from pp.components.extension import extend_ports
from pp.add_padding import add_padding
from pp.add_tapers import add_tapers
from pp.rotate import rotate
from pp.add_grating_couplers import add_grating_couplers
from pp.add_termination import add_termination, add_gratings_and_loop_back
from pp.components.spiral_inner_io import spiral_inner_io

# from pp.components.waveguide import waveguide
# from pp.components.waveguide_heater import waveguide_heater
from pp.components import mzi2x2

from pp.routing import add_fiber_array
from pp.routing import add_fiber_single
from pp.routing import add_electrical_pads
from pp.routing import add_electrical_pads_top
from pp.routing import add_electrical_pads_shortest
from pp.routing import package_optical2x2


_containers = [
    extend_ports,
    add_padding,
    add_tapers,
    rotate,
    add_termination,
    add_fiber_single,
    add_fiber_array,
    add_electrical_pads,
    add_electrical_pads_top,
    add_electrical_pads_shortest,
    add_grating_couplers,
    package_optical2x2,
]


def test_add_gratings_and_loop_back(data_regression):
    c = add_gratings_and_loop_back(component=spiral_inner_io())
    data_regression.check(c.get_settings())


@pytest.mark.parametrize("function", _containers)
def test_properties_containers(function, data_regression):
    component = mzi2x2(with_elec_connections=True)
    c = function(component=component)
    data_regression.check(c.get_settings())



"""
Lets define the references from a component and then connect them together

"""

import pp


@pp.autoname
def test_ring_single_bus(
    coupler90_factory=pp.c.coupler90,
    cpl_straight_factory=pp.c.coupler_straight,
    straight_factory=pp.c.waveguide,
    bend90_factory=pp.c.bend_circular,
    length_y=2.0,
    length_x=4.0,
    gap=0.2,
    wg_width=0.5,
    bend_radius=5,
):
    """ single bus ring
    """
    c = pp.Component()

    # define subcells
    coupler90 = pp.call_if_func(
        coupler90_factory, gap=gap, width=wg_width, bend_radius=bend_radius
    )
    waveguide_x = pp.call_if_func(straight_factory, length=length_x, width=wg_width)
    waveguide_y = pp.call_if_func(straight_factory, length=length_y, width=wg_width)
    bend = pp.call_if_func(bend90_factory, width=wg_width, radius=bend_radius)
    coupler_straight = pp.call_if_func(
        cpl_straight_factory, gap=gap, length=length_x, width=wg_width
    )

    # add references to subcells
    cbl = c << coupler90
    cbr = c << coupler90
    cs = c << coupler_straight
    wyl = c << waveguide_y
    wyr = c << waveguide_y
    wx = c << waveguide_x
    btl = c << bend
    btr = c << bend

    # connect references
    wyr.connect(port="E0", destination=cbr.ports["N0"])
    cs.connect(port="E0", destination=cbr.ports["W0"])

    cbl.reflect(p1=(0, coupler90.y), p2=(1, coupler90.y))
    cbl.connect(port="W0", destination=cs.ports["W0"])
    wyl.connect(port="E0", destination=cbl.ports["N0"])

    btl.connect(port="N0", destination=wyl.ports["W0"])
    btr.connect(port="W0", destination=wyr.ports["W0"])
    wx.connect(port="W0", destination=btl.ports["W0"])

    c.add_port("W0", port=cbl.ports["E0"])
    c.add_port("E0", port=cbr.ports["E0"])
    assert c
    return c


if __name__ == "__main__":
    c = test_ring_single_bus(wg_width=0.45, gap=0.15, length_x=0.2, length_y=0.13)
    pp.show(c)
    # pp.write_gds(c, "ring.gds")



import pp
from pp.components.bend_circular import bend_circular
from pp.components.coupler90 import coupler90
from pp.components.waveguide import waveguide
from pp.components.coupler_straight import coupler_straight
from pp.components.taper import taper
from pp.netlist_to_gds import netlist_to_component
from pp.name import autoname
from pp.drc import assert_on_2nm_grid


@autoname
def ring_with_taper(**kwargs):
    """ Ring single bus

    Args:
        bend_radius=5
        length_x=1
        length_y=1
        gap=0.2
        taper_factory=taper
        bend90_factory=bend_circular
        coupler90_factory=coupler90
        straight_factory=waveguide
        cpl_straight_factory=coupler_straight
        wg_width=0.5


    .. code::

         --length_x-
         |         |
         |       length_y
         |         |
       ---===gap===---

    .. plot::
      :include-source:

      import pp

      c = pp.c.ring_single_bus(gap=0.2, length_x=10, length_y=5, bend_radius=5)
      pp.plotgds(c)

    """
    components, connections, ports_map = ring_with_taper_netlist(**kwargs)
    component = netlist_to_component(components, connections, ports_map)
    return component


def ring_with_taper_netlist(
    bend_radius=5,
    length_x=1,
    length_y=0,
    gap=0.2,
    taper_width=1.0,
    taper_length=10,
    taper_factory=taper,
    bend90_factory=bend_circular,
    coupler90_factory=coupler90,
    straight_factory=waveguide,
    cpl_straight_factory=coupler_straight,
    wg_width=0.5,
):
    """
    .. code::

         BL--H1--BR
         |       |
         |       T2
         |       VR length_y
         VL      T1
         |       |
        -CL==CS==CR-
    """
    assert_on_2nm_grid(gap)

    taper = taper_factory(length=taper_length, width1=wg_width, width2=taper_width)
    bend = bend90_factory(radius=bend_radius, width=wg_width)
    cpl_bend = coupler90_factory(bend_radius=bend_radius, width=wg_width, gap=gap)
    cpl_straight = cpl_straight_factory(length=length_x, gap=gap, width=wg_width)
    h1 = straight_factory(length=length_x, width=wg_width)
    vl = straight_factory(length=length_y + 2 * taper_length, width=wg_width)
    vr = straight_factory(length=length_y, width=taper_width)

    components = {
        "CL": (cpl_bend, "mirror_y"),
        "CR": (cpl_bend, "None"),
        "BR": (bend, "mirror_x"),
        "BL": (bend, "R180"),
        "CS": (cpl_straight, "None"),
        "H1": (h1, "None"),
        "T1": (taper, "R90"),
        "T2": (taper, "R270"),
        "VL": (vl, "R90"),
        "VR": (vr, "R90"),
    }
    connections = [
        ("CL", "W0", "CS", "W0"),
        ("CS", "E0", "CR", "W0"),
        ("CR", "N0", "T1", "1"),
        ("T1", "2", "VR", "W0"),
        ("VR", "E0", "T2", "2"),
        ("T2", "1", "BR", "N0"),
        ("BR", "W0", "H1", "E0"),
        ("H1", "W0", "BL", "W0"),
        ("BL", "N0", "VL", "E0"),
        ("VL", "W0", "CL", "N0"),
    ]

    ports_map = {"W0": ("CL", "E0"), "E0": ("CR", "E0")}

    return components, connections, ports_map


if __name__ == "__main__":
    # c = ring_with_taper(bend_radius=5.0, length_x=2, length_y=4, gap=0.2)
    c = ring_with_taper(
        bend_radius=5.0, length_x=2, length_y=4, gap=0.2, taper_width=1, wg_width=0.7
    )
    pp.show(c)



"""
load component gds, and ports
"""
import os
import pathlib
import json
import csv

from pp import CONFIG
import pp
from pathlib import PosixPath
from pp.component import Component


def get_component_path(component_name, component_path=CONFIG["gdslib"]):
    return component_path / (component_name + ".gds")


def load_component_path(component_name, component_path=CONFIG["gdslib"]):
    """ load component GDS from a library
    returns a gdspath
    """
    gdspath = component_path / (component_name + ".gds")

    if not os.path.isfile(gdspath):
        raise ValueError("cannot load `{}`".format(gdspath))

    return gdspath


def load_component(
    component_name: str,
    component_path: PosixPath = CONFIG["gdslib"],
    with_info_labels: bool = True,
    overwrite_cache: bool = False,
) -> Component:
    """ loads GDS, ports (CSV) and metadata (JSON)
    returns a Device

    Args:
        component_name:
        component_path: libary path
        with_info_labels: can remove labal info
        overwrite_cache
    """
    component_path = pathlib.Path(component_path)

    gdspath = component_path / (component_name + ".gds")
    portspath = gdspath.with_suffix(".ports")
    jsonpath = gdspath.with_suffix(".json")

    if not os.path.isfile(gdspath):
        raise ValueError("cannot load `{}`".format(gdspath))

    c = pp.import_gds(str(gdspath), overwrite_cache=overwrite_cache)
    c.name = component_name

    # Remove info labels if needed
    if not with_info_labels:
        for component in list(c.get_dependencies(recursive=True)) + [c]:
            old_label = [
                label
                for label in component.labels
                if label.layer == pp.LAYER.INFO_GEO_HASH
            ]
            if len(old_label) > 0:
                for label in old_label:
                    component.labels.remove(label)

    """ add ports """
    try:
        with open(str(portspath), newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=",", quotechar="|")
            for r in reader:
                layer_type = int(r[5].strip().strip("("))
                data_type = int(r[6].strip().strip(")"))
                c.add_port(
                    name=r[0],
                    midpoint=[float(r[1]), float(r[2])],
                    orientation=int(r[3]),
                    width=float(r[4]),
                    layer=(layer_type, data_type),
                )
    except Exception:
        # print(
        #     f"Could not find a port CSV file for {component_name} in {portspath}"
        # )
        # print(
        #     "ports follow (name, x, y, width, angle, layer_gds_type, layer_gds_purpose)"
        # )
        pass
        # raise (e)

    """ add settings """
    try:
        with open(jsonpath) as f:
            data = json.load(f)
        cell_settings = data["cells"][c.name]
        c.settings.update(cell_settings)
    except Exception:
        pass
        print(f"could not load settings for {c.name} in {jsonpath}")
    return c


def _compare_hash():
    component_type = "coupler90"
    c = load_component(component_type)
    c2 = pp.get_component_type(component_type, gap=0.1)
    print(c.hash_geometry())
    print(c2.hash_geometry())
    pp.show(c)


if __name__ == "__main__":
    component_type = "waveguide"
    c = load_component(component_type)
    print(c.settings)



from typing import Callable
from typing import Any, List, Optional, Tuple, Dict, Union
import functools
from copy import deepcopy
import csv
import numpy as np
import phidl.geometry as pg
from phidl.device_layout import Port as PortPhidl
from phidl.device_layout import Device
from pp.drc import snap_to_grid

port_types = ["optical", "rf", "dc", "heater"]


class Port(PortPhidl):
    """Extends phidl port with layer and port_type (optical, dc, rf)

    Args:
        name: we name ports according to orientation (S0, S1, W0, W1, N0 ...)
        midpoint: (0, 0)
        width: of the port
        orientation: 0
        parent: parent component (component to which this port belong to)
        layer: (1, 0)
        port_type: optical, dc, rf, detector, superconducting, trench

    """

    _next_uid = 0

    def __init__(
        self,
        name: Optional[str] = None,
        midpoint: Tuple[float, float] = (0.0, 0.0),
        width: float = 0.5,
        orientation: int = 0,
        parent: Optional[object] = None,
        layer: Tuple[int, int] = (1, 0),
        port_type: str = "optical",
    ) -> None:
        self.name = name
        self.midpoint = np.array(midpoint, dtype="float64")
        self.width = width
        self.orientation = np.mod(orientation, 360)
        self.parent = parent
        self.info = {}
        self.uid = Port._next_uid
        self.layer = layer
        self.port_type = port_type

        if self.width < 0:
            raise ValueError("[PHIDL] Port creation error: width must be >=0")
        self._next_uid += 1

    def __repr__(self) -> str:
        return (
            "Port (name {}, midpoint {}, width {}, orientation {}, layer {},"
            " port_type {})".format(
                self.name,
                self.midpoint,
                self.width,
                self.orientation,
                self.layer,
                self.port_type,
            )
        )

    @property
    def angle(self):
        """convenient alias"""
        return self.orientation

    @angle.setter
    def angle(self, a):
        self.orientation = a

    @property
    def position(self):
        return self.midpoint

    @position.setter
    def position(self, p):
        self.midpoint = np.array(p, dtype="float64")

    def move(self, vector):
        self.midpoint = self.midpoint + np.array(vector)

    def move_polar_copy(self, d, angle):
        port = self._copy()
        DEG2RAD = np.pi / 180
        dp = np.array((d * np.cos(DEG2RAD * angle), d * np.sin(DEG2RAD * angle)))
        self.move(dp)
        return port

    def flip(self):
        """ flips port """
        port = self._copy()
        port.angle = (port.angle + 180) % 360
        return port

    def _copy(self, new_uid: bool = True):
        new_port = Port(
            name=self.name,
            midpoint=self.midpoint,
            width=self.width,
            orientation=self.orientation,
            parent=self.parent,
            layer=self.layer,
            port_type=self.port_type,
        )
        new_port.info = deepcopy(self.info)
        if not new_uid:
            new_port.uid = self.uid
            Port._next_uid -= 1
        return new_port

    def snap_to_grid(self, nm=1):
        self.midpoint = nm * np.round(np.array(self.midpoint) * 1e3 / nm) / 1e3

    def on_grid(self, nm: int = 1) -> None:
        if self.orientation in [0, 180]:
            x = self.y + self.width / 2
            assert np.isclose(
                snap_to_grid(x, nm=nm), x
            ), f"{self.parent} port {self.name} has an off-grid point {x}"
        elif self.orientation in [90, 270]:
            x = self.x + self.width / 2
            assert np.isclose(
                snap_to_grid(x, nm=nm), x
            ), f"{self.parent} port {self.name} has an off-grid point {x}"
        else:
            raise ValueError(
                f"{self.parent} port {self.name} has invalid orientation"
                f" {self.orientation}"
            )


def read_port_markers(gdspath, layers=[(69, 0)]):
    """loads a GDS and returns the extracted device for a particular layer

    Args:
        gdspath: gdspath or Component
        layer: GDS layer
    """
    D = gdspath if isinstance(gdspath, Device) else pg.import_gds(gdspath)
    return pg.extract(D, layers=layers)


def csv2port(csvpath):
    """loads and reads ports from a CSV file
    returns a dict
    """
    ports = {}
    with open(csvpath, "r") as csvfile:
        rows = csv.reader(csvfile, delimiter=",", quotechar="|")
        for row in rows:
            ports[row[0]] = row[1:]

    return ports


def is_electrical_port(port):
    return port.port_type in ["dc", "rf"]


def select_ports(
    ports,
    port_type: Union[str, Tuple[int, int]] = "optical",
    prefix: Optional[str] = None,
):
    """
    Args:
        ports: Dict[str, Port] a port dictionnary {port name: port} (as returned by Component.ports)
        layers: a list of port layer or a port type (layer or string)

    Returns:
        Dictionnary containing only the ports with the wanted type(s)
        {port name: port}
    """

    from pp.component import Component, ComponentReference

    # Make it accept Component or ComponentReference
    if isinstance(ports, Component) or isinstance(ports, ComponentReference):
        ports = ports.ports

    ports = {
        p_name: p
        for p_name, p in ports.items()
        if p.port_type == port_type or p.layer == port_type
    }
    if prefix:
        ports = {p_name: p for p_name, p in ports.items() if p_name.startswith(prefix)}
    return ports


def select_optical_ports(ports: Dict[str, Port], prefix=None) -> Dict[str, Port]:
    return select_ports(ports, port_type="optical", prefix=prefix)


def select_electrical_ports(ports, port_type="dc", prefix=None):
    d = select_ports(ports, port_type=port_type, prefix=prefix)
    d.update(select_ports(ports, port_type="electrical"))
    return d


def flipped(port):
    _port = port._copy()
    _port.orientation = (_port.orientation + 180) % 360
    return _port


def move_copy(port, x=0, y=0):
    _port = port._copy()
    _port.midpoint += (x, y)
    return _port


def get_ports_facing(ports, direction="W"):
    from pp.component import Component, ComponentReference

    if isinstance(ports, dict):
        ports = list(ports.values())
    elif isinstance(ports, Component) or isinstance(ports, ComponentReference):
        ports = list(ports.ports.values())

    direction_ports = {x: [] for x in ["E", "N", "W", "S"]}

    for p in ports:
        angle = p.orientation % 360
        if angle <= 45 or angle >= 315:
            direction_ports["E"].append(p)
        elif angle <= 135 and angle >= 45:
            direction_ports["N"].append(p)
        elif angle <= 225 and angle >= 135:
            direction_ports["W"].append(p)
        else:
            direction_ports["S"].append(p)

    return direction_ports[direction]


def get_non_optical_ports(ports):
    from pp.component import Component, ComponentReference

    if isinstance(ports, dict):
        ports = list(ports.values())
    elif isinstance(ports, Component) or isinstance(ports, ComponentReference):
        ports = list(ports.ports.values())
    res = [p for p in ports if p.port_type not in ["optical"]]
    return res


def deco_rename_ports(component_factory: Callable) -> Callable:
    @functools.wraps(component_factory)
    def auto_named_component_factory(*args, **kwargs):
        component = component_factory(*args, **kwargs)
        auto_rename_ports(component)
        return component

    return auto_named_component_factory


def _rename_ports_facing_side(
    direction_ports: Dict[str, List[Port]], prefix: str = ""
) -> None:
    for direction, list_ports in list(direction_ports.items()):

        if direction in ["E", "W"]:
            # first sort along x then y
            list_ports.sort(key=lambda p: p.x)
            list_ports.sort(key=lambda p: p.y)

        if direction in ["S", "N"]:
            # first sort along y then x
            list_ports.sort(key=lambda p: p.y)
            list_ports.sort(key=lambda p: p.x)

        for i, p in enumerate(list_ports):
            lbl = prefix + direction + str(i)
            p.name = lbl


def rename_ports_by_orientation(
    component: object, layers_excluded: List[Any] = []
) -> object:
    """Returns Component with port names based on port orientation (E, N, W, S)"""

    direction_ports = {x: [] for x in ["E", "N", "W", "S"]}
    ports_on_process = [
        p for p in component.ports.values() if p.layer not in layers_excluded
    ]

    for p in ports_on_process:
        # Make sure we can backtrack the parent component from the port
        p.parent = component

        angle = p.orientation % 360
        if angle <= 45 or angle >= 315:
            direction_ports["E"].append(p)
        elif angle <= 135 and angle >= 45:
            direction_ports["N"].append(p)
        elif angle <= 225 and angle >= 135:
            direction_ports["W"].append(p)
        else:
            direction_ports["S"].append(p)

    _rename_ports_facing_side(direction_ports)
    component.ports = {p.name: p for p in component.ports.values()}
    return component


def auto_rename_ports(component: object) -> object:
    """Returns Component with port names based on port orientation (E, N, W, S)"""

    def _counter_clockwise(_direction_ports, prefix=""):

        east_ports = _direction_ports["E"]
        east_ports.sort(key=lambda p: p.y)  # sort south to north

        north_ports = _direction_ports["N"]
        north_ports.sort(key=lambda p: -p.x)  # sort east to west

        west_ports = _direction_ports["W"]
        west_ports.sort(key=lambda p: -p.y)  # sort north to south

        south_ports = _direction_ports["S"]
        south_ports.sort(key=lambda p: p.x)  # sort west to east

        ports = east_ports + north_ports + west_ports + south_ports

        for i, p in enumerate(ports):
            p.name = "{}{}".format(prefix, i)

    type_to_ports_naming_functions = {
        "optical": _rename_ports_facing_side,
        "heater": lambda _d: _counter_clockwise(_d, "H_"),
        "dc": lambda _d: _counter_clockwise(_d, "E_"),
        "superconducting": lambda _d: _counter_clockwise(_d, "SC_"),
    }

    type_to_ports = {}

    for p in component.ports.values():
        if p.port_type not in type_to_ports:
            type_to_ports[p.port_type] = []
        type_to_ports[p.port_type] += [p]

    for port_type, port_group in type_to_ports.items():
        if port_type in type_to_ports_naming_functions:
            _func_name_ports = type_to_ports_naming_functions[port_type]
        else:
            raise ValueError(
                "Unknown port type <{}> in component {}, port {}".format(
                    port_type, component.name, p
                )
            )

        # Make sure we can backtrack the parent component from the port

        direction_ports = {x: [] for x in ["E", "N", "W", "S"]}
        for p in port_group:
            p.parent = component
            angle = p.orientation % 360
            if angle <= 45 or angle >= 315:
                direction_ports["E"].append(p)
            elif angle <= 135 and angle >= 45:
                direction_ports["N"].append(p)
            elif angle <= 225 and angle >= 135:
                direction_ports["W"].append(p)
            else:
                direction_ports["S"].append(p)

        _func_name_ports(direction_ports)

    # Set the port dictionnary with the new names
    component.ports = {p.name: p for p in component.ports.values()}
    return component


def test_select_ports_prefix():
    import pp

    c = pp.c.waveguide()
    ports = c.get_ports_list(prefix="W")
    assert len(ports) == 1


def test_select_ports_type():
    import pp

    c = pp.c.mzi2x2(with_elec_connections=True)
    ports = c.get_ports_list(port_type="dc")
    assert len(ports) == 3


if __name__ == "__main__":
    test_select_ports_type()

    import pp

    name = "mmi1x2"
    gdspath = pp.CONFIG["gdslib"] / "gds" / f"{name}.gds"
    csvpath = pp.CONFIG["gdslib"] / "gds" / f"{name}.ports"
    pp.show(gdspath)
    # read_port_markers(gdspath, layer=66)
    p = csv2port(csvpath)
    print(p)



import pp


def test_label_move():
    """ test that when we move a device its label also moves """
    c = pp.Component("ellipse_with_label")
    c << pp.c.ellipse()
    c.add_label(text="demo", position=(10, 0), layer=pp.LAYER.TEXT)
    c.movex(10)
    print(c.references)
    print(c.labels)
    # assert c.references[0].origin[0] == 10
    # assert c.labels[0].position[0] == 20
    return c


if __name__ == "__main__":
    c = test_label_move()
    pp.show(c)



import itertools
import pathlib
import gdspy as gp

from pp import import_gds
from pp.layers import get_gds_layers
import pp

COUNTER = itertools.count()


def boolean(A, B, operation, precision):
    p = gp.boolean(
        operand1=A,
        operand2=B,
        operation=operation,
        precision=precision,
        max_points=4000,
    )
    return p


def get_polygons_on_layer(cell, layer):
    polygons = cell.get_polygons(by_spec=True)
    # pprint ({k: len(v) for k, v in polygons.items()})
    if layer in polygons:
        return polygons[layer]
    else:
        return None


def gdsdiff(cellA, cellB):
    """
    Args:
        CellA: gds cell (as pp.Component) or path to gds file
        CellB: gds cell (as pp.Component) or path to gds file

    Output:
        gds file containing the diff between the two GDS files
    """
    if isinstance(cellA, pathlib.PosixPath):
        cellA = str(cellA)
    if isinstance(cellB, pathlib.PosixPath):
        cellB = str(cellB)
    if type(cellA) == str:
        cellA = import_gds(cellA, flatten=True)
    if type(cellB) == str:
        cellB = import_gds(cellB, flatten=True)

    layers = set()
    layers.update(get_gds_layers(cellA))
    layers.update(get_gds_layers(cellB))

    diff = pp.Component(name="diff")
    for layer in layers:
        # We go to "process" layer beyond 1000 to put the diff
        diff_process = layer[0] + 1000

        """
        # datatype is used to differentiate the diff
        # 0: unchanged
        # 1: removed (assuming B is the updated version of A)
        # 2: added (assuming B is the updated version of A)
        """

        layer_common = (diff_process, 0)
        layer_only_A = (diff_process, 1)
        layer_only_B = (diff_process, 2)

        A = get_polygons_on_layer(cellA, layer)
        B = get_polygons_on_layer(cellB, layer)

        if A is None and B is None:
            continue
        elif B is None:
            diff.add_polygon(A, layer_only_A)
            continue
        elif A is None:
            diff.add_polygon(B, layer_only_B)
            continue

        # Common bits
        common_AB = boolean(A, B, operation="and", precision=0.001)

        # Bits in either A or B
        either_AB = boolean(A, B, operation="xor", precision=0.001)

        # Bits only in A
        only_in_A = boolean(A, either_AB, operation="and", precision=0.001)

        # Bits only in B
        only_in_B = boolean(B, either_AB, operation="and", precision=0.001)

        if common_AB is not None:
            diff.add_polygon(common_AB, layer_common)
        if only_in_A is not None:
            diff.add_polygon(only_in_A, layer_only_A)
        if only_in_B is not None:
            diff.add_polygon(only_in_B, layer_only_B)

    return diff


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 3:
        print("Usage: gdsdiff <mask_v1.gds> <mask_v2.gds>")
        print("Note that you need to have KLayout opened with klive running")
        sys.exit()

    diff = gdsdiff(sys.argv[1], sys.argv[2])
    pp.show(diff)



from typing import Tuple
import numpy as np
from numpy import pi, cos, sin

import pp
from pp.component import Component


@pp.autoname
def ring(
    radius: float = 10.0,
    width: float = 0.5,
    angle_resolution: float = 2.5,
    layer: Tuple[int, int] = pp.LAYER.WG,
) -> Component:
    """ Returns ring geometry.
    The ring is formed by taking the radius out to the specified value, and then constructing the thickness by dividing the width in half and adding that value to either side of the radius.
    The angle_resolution alters the precision of the curve of the ring.
    Larger values yield lower resolution

    Args:
        radius: (float) Middle radius of the ring
        width: (float) Width of the ring
        angle_resolution: (float) Resolution of the curve of the ring (# of degrees per point)
        layer: (int, array-like[2], or set) Specific layer(s) to put polygon geometry on

    .. plot::
      :include-source:

      import pp

      c = pp.c.ring(radius=10, width=0.5, angle_resolution=2.5, layer=0)
      pp.plotgds(c)

    """

    D = pp.Component()
    inner_radius = radius - width / 2
    outer_radius = radius + width / 2
    n = int(np.round(360 / angle_resolution))
    t = np.linspace(0, 360, n + 1) * pi / 180
    inner_points_x = (inner_radius * cos(t)).tolist()
    inner_points_y = (inner_radius * sin(t)).tolist()
    outer_points_x = (outer_radius * cos(t)).tolist()
    outer_points_y = (outer_radius * sin(t)).tolist()
    xpts = inner_points_x + outer_points_x[::-1]
    ypts = inner_points_y + outer_points_y[::-1]
    D.add_polygon(points=(xpts, ypts), layer=layer)
    return D


if __name__ == "__main__":
    c = ring()
    pp.show(c)



import numpy as np
import picwriter.components as pc

from pp.components.waveguide_template import wg_strip
from pp.picwriter2component import picwriter2component
import pp


@pp.autoname
def grating_coupler_elliptical2(
    wgt=wg_strip,
    theta=np.pi / 4.0,
    length=30.0,
    taper_length=10.0,
    period=1.0,
    dutycycle=0.7,
    ridge=True,
    ridge_layers=(2, 0),
    teeth_list=None,
    port=(0, 0),
    direction="EAST",
    **kwargs
):
    """ Grating coupler

    Args:
        waveguide_template: object or function
        port (tuple): Cartesian coordinate of the input port
        direction (string): Direction that the component will point *towards*, can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`, OR an angle (float, in radians)
        theta (float): Angle of the waveguide.  Defaults to pi/4.0
        length (float): Length of the total grating coupler region, measured from the output port.  Defaults to 30.0
        taper_length (float): Length of the taper before the grating coupler.  Defaults to 10.0
        period (float): Grating period.  Defaults to 1.0
        dutycycle (float): dutycycle, determines the size of the 'gap' by dutycycle=(period-gap)/period.  Defaults to 0.7
        ridge (boolean): If True, adds another layer to the grating coupler that can be used for partial etched gratings
        ridge_layers (tuple): Tuple specifying the layer/datatype of the ridge region.  Defaults to (3,0)
        teeth_list (list): Can optionally pass a list of (gap, width) tuples to be used as the gap and teeth widths for irregularly spaced gratings.  For example, [(0.6, 0.2), (0.7, 0.3), ...] would be a gap of 0.6, then a tooth of width 0.2, then gap of 0.7 and tooth of 0.3, and so on.  Overrides *period*, *dutycycle*, and *length*.  Defaults to None.

    .. plot::
      :include-source:

      import pp

      c = pp.c.grating_coupler_elliptical2()
      pp.plotgds(c)

    """

    c = pc.GratingCoupler(
        pp.call_if_func(wg_strip, **kwargs),
        theta=theta,
        length=length,
        taper_length=taper_length,
        period=period,
        dutycycle=dutycycle,
        ridge=ridge,
        ridge_layers=ridge_layers,
        teeth_list=teeth_list,
        port=port,
        direction=direction,
    )

    return picwriter2component(c)


if __name__ == "__main__":
    import pp

    c = grating_coupler_elliptical2(pins=True)
    pp.show(c)



import picwriter.components as pc
import pp
from picwriter.components.waveguide import WaveguideTemplate


def wg_strip(
    wg_width: float = 0.5,
    wg_layer: int = pp.LAYER.WG[0],
    wg_datatype: int = pp.LAYER.WG[1],
    clad_layer: int = pp.LAYER.WGCLAD[0],
    clad_datatype: int = pp.LAYER.WGCLAD[1],
    bend_radius: int = 10,
    cladding_offset: int = 3,
) -> WaveguideTemplate:
    return pc.WaveguideTemplate(
        bend_radius=bend_radius,
        wg_width=wg_width,
        wg_layer=wg_layer,
        wg_datatype=wg_datatype,
        clad_layer=clad_layer,
        clad_datatype=clad_datatype,
        clad_width=cladding_offset,
    )



from click.testing import CliRunner
from pp.pf import cli
from pp import __version__


def test_pf_version():
    """ checks that the CLI returns the correct version """
    runner = CliRunner()
    result = runner.invoke(cli, ["config", "version"])

    print(result.output)
    print(__version__)
    assert result.exit_code == 0
    assert result.output.startswith(__version__)


if __name__ == "__main__":
    test_pf_version()



"""
From two GDS files, find which cells are identical
and which cells with the same name are different
This is not a diff tool
"""
import hashlib
import sys
import json
import numpy as np


def _print(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()


def get_transform(cell_ref, precision=1e-4):
    """
    Get the transform from a cell-instance as a hashable object
    """
    if cell_ref.rotation is None:
        cell_ref.rotation = 0
    return (
        int(cell_ref.origin[0] / precision),
        int(cell_ref.origin[1] / precision),
        int(cell_ref.rotation) % 360,
        cell_ref.x_reflection,
    )


def get_polygons_by_spec(cell):
    d = {}
    for _pset in cell.polygons:
        for poly, layer, datatype in zip(_pset.polygons, _pset.layers, _pset.datatypes):
            key = (layer, datatype)
            if key not in d:
                d[key] = []
            d[key].append(poly)

    return d


def get_dependencies_names(cell):
    return [_c.ref_cell.name for _c in cell.references]


def normalize_polygon_start_point(p, dbg=False):
    args_min_x = np.where(p[:, 0] == p[:, 0].min())[0]
    if len(args_min_x) == 1:
        i0 = args_min_x[0]
    else:
        ys = np.array([p[i, 1] for i in args_min_x])
        i0 = args_min_x[np.argmin(ys)]

    if dbg:
        _print(p)
        _print(i0, args_min_x)
        _print()

    return np.roll(p, -i0, axis=0)


def hash_cells(cell, dict_hashes={}, precision=1e-4, dbg_indent=0, dbg=False):
    """
    Algorithm:
    For each polygon directly within this cell:
         - sort the layers

        For each layer, each polygon is individually hashed and then
          the polygon hashes are sorted, to ensure the hash stays constant
          regardless of the ordering the polygons.  Similarly, the layers
          are sorted by (layer, datatype)

    For each cell instance:
        recursively hash the ref_cell + transform
        sort all the hashes for the hash to stay constant regardless of cell instance order

    """
    if cell.name in dict_hashes:
        return dict_hashes

    polygons_by_spec = get_polygons_by_spec(cell)
    layers = list(polygons_by_spec.keys())
    layers.sort()

    if dbg:
        _print(layers)

    """
    # A random offset which fixes common rounding errors intrinsic
    # to floating point math. Example: with a precision of 0.1, the
    # floating points 7.049999 and 7.050001 round to different values
    # (7.0 and 7.1), but offset values (7.220485 and 7.220487) don't
    """
    magic_offset = 0.17048614

    final_hash = hashlib.sha1()

    for layer in layers:
        layer_hash = hashlib.sha1(str(layer).encode()).hexdigest()
        polygons = polygons_by_spec[tuple(layer)]

        polygons = [((p / precision) + magic_offset).astype(np.int64) for p in polygons]
        polygons = [normalize_polygon_start_point(p) for p in polygons]
        polygon_hashes = np.sort([hashlib.sha1(p).hexdigest() for p in polygons])

        if dbg:
            _print(polygons)
            _print(layer, layer_hash, polygon_hashes)

        final_hash.update(layer_hash.encode())
        for ph in polygon_hashes:
            final_hash.update(ph.encode())

    # Ref cell hashes
    cell_ref_uids = []
    for cell_ref in cell.references:
        _cell = cell_ref.ref_cell
        hash_cells(
            _cell,
            precision=precision,
            dict_hashes=dict_hashes,
            dbg_indent=dbg_indent + 2,
            dbg=dbg,
        )
        cell_hash = dict_hashes[_cell.name]
        tr_str = "x{}y{}R{}H{}".format(*get_transform(cell_ref, precision))

        cell_ref_uid = cell_hash + "_" + tr_str
        cell_ref_uids += [cell_ref_uid]

    # Sort hashes (for constant hash regardless of cell_ref ordering)
    cell_ref_uids = np.sort(cell_ref_uids)
    for _hash in cell_ref_uids:
        final_hash.update(_hash.encode())

    dict_hashes[cell.name] = final_hash.hexdigest()
    if dbg:
        _print()
    return dict_hashes


def check_lib_consistency(lib_cells):
    """
    TODO: check that the library does not have a name collision (with different hashes)
    """


def is_leaf_cell(cell):
    return len(cell.references) == 0


def get_cell_status(cell, name_to_hash_teg, name_to_hash_lib, cells_status={}):
    if cell.name in cells_status:
        return cells_status[cell.name]

    sub_cells_names = []
    if cell.name in name_to_hash_lib:
        """
        Known leaf cell: it ether matches the lib or not
        """

        # _print("Known",cell.name , name_to_hash_teg[cell.name][:8], name_to_hash_lib[cell.name][:8])

        if name_to_hash_teg[cell.name] == name_to_hash_lib[cell.name]:
            cell_status = 0
        else:
            cell_status = 2

    elif is_leaf_cell(cell):
        """
        Unknown leaf cell
        """
        cell_status = 1

    else:
        """
        Compound cell not in lib: inherit status from subcells
        """
        sub_cells = [_c.ref_cell for _c in cell.references]
        cell_status = max(
            [
                get_cell_status(_c, name_to_hash_teg, name_to_hash_lib, cells_status)[0]
                for _c in sub_cells
            ]
        )
        sub_cells_names = [_c.name for _c in sub_cells]

    cell_status_and_dependencies = (cell_status, sub_cells_names)
    cells_status[cell.name] = cell_status_and_dependencies
    return cell_status_and_dependencies


def compare_gds_to_lib(teg, lib_cells=[]):
    """
    teg: a cell
    lib_cell: list of cells

    output:
        dictionnary {cell_name: status}
        where status is:
            0 if the cell matches the library, or is composed of cells with status 0
            1 if the cell does not match any lib cell of at least one cell with status 1 and all the others with status 0
            2 if the cell name is in the library but the hashes do not match, OR if the cell is composed of at least one cell with status 2
    """

    name_to_hash_teg = {}
    hash_cells(teg, name_to_hash_teg)
    ds_lib = [hash_cells(_c, {}) for _c in lib_cells]

    check_lib_consistency(ds_lib)
    name_to_hash_lib = {}
    while ds_lib:
        name_to_hash_lib.update(ds_lib.pop())

    with open("dbg.json", "w") as fw:
        fw.write(
            json.dumps({"LIB": name_to_hash_lib, "TEG": name_to_hash_teg}, indent=2)
        )

    # hash_to_name_teg = {v: k for k, v in name_to_hash_teg.items()}
    # hash_to_name_lib = {v: k for k, v in name_to_hash_lib.items()}

    cells_status = {}
    get_cell_status(teg, name_to_hash_teg, name_to_hash_lib, cells_status)
    return cells_status


if __name__ == "__main__":
    pass



import pp
from pp import Port
from pp.routing.connect_bundle import connect_bundle


def test_connect_bundle():

    xs_top = [-100, -90, -80, 0, 10, 20, 40, 50, 80, 90, 100, 105, 110, 115]

    pitch = 127.0
    N = len(xs_top)
    xs_bottom = [(i - N / 2) * pitch for i in range(N)]

    top_ports = [Port("top_{}".format(i), (xs_top[i], 0), 0.5, 270) for i in range(N)]

    bottom_ports = [
        Port("bottom_{}".format(i), (xs_bottom[i], -400), 0.5, 90) for i in range(N)
    ]

    top_cell = pp.Component(name="connect_bundle")
    elements = connect_bundle(top_ports, bottom_ports)
    for e in elements:
        top_cell.add(e)
    top_cell.name = "connect_bundle"

    return top_cell


@pp.autoname
def test_connect_corner(N=6, config="A"):

    d = 10.0

    sep = 5.0
    top_cell = pp.Component(name="connect_bundle_corners")

    if config in ["A", "B"]:
        a = 100.0
        ports_A_TR = [
            Port("A_TR_{}".format(i), (d, a / 2 + i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_TL = [
            Port("A_TL_{}".format(i), (-d, a / 2 + i * sep), 0.5, 180) for i in range(N)
        ]
        ports_A_BR = [
            Port("A_BR_{}".format(i), (d, -a / 2 - i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_BL = [
            Port("A_BL_{}".format(i), (-d, -a / 2 - i * sep), 0.5, 180)
            for i in range(N)
        ]

        ports_A = [ports_A_TR, ports_A_TL, ports_A_BR, ports_A_BL]

        ports_B_TR = [
            Port("B_TR_{}".format(i), (a / 2 + i * sep, d), 0.5, 90) for i in range(N)
        ]
        ports_B_TL = [
            Port("B_TL_{}".format(i), (-a / 2 - i * sep, d), 0.5, 90) for i in range(N)
        ]
        ports_B_BR = [
            Port("B_BR_{}".format(i), (a / 2 + i * sep, -d), 0.5, 270) for i in range(N)
        ]
        ports_B_BL = [
            Port("B_BL_{}".format(i), (-a / 2 - i * sep, -d), 0.5, 270)
            for i in range(N)
        ]

        ports_B = [ports_B_TR, ports_B_TL, ports_B_BR, ports_B_BL]

    elif config in ["C", "D"]:
        a = N * sep + 2 * d
        ports_A_TR = [
            Port("A_TR_{}".format(i), (a, d + i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_TL = [
            Port("A_TL_{}".format(i), (-a, d + i * sep), 0.5, 180) for i in range(N)
        ]
        ports_A_BR = [
            Port("A_BR_{}".format(i), (a, -d - i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_BL = [
            Port("A_BL_{}".format(i), (-a, -d - i * sep), 0.5, 180) for i in range(N)
        ]

        ports_A = [ports_A_TR, ports_A_TL, ports_A_BR, ports_A_BL]

        ports_B_TR = [
            Port("B_TR_{}".format(i), (d + i * sep, a), 0.5, 90) for i in range(N)
        ]
        ports_B_TL = [
            Port("B_TL_{}".format(i), (-d - i * sep, a), 0.5, 90) for i in range(N)
        ]
        ports_B_BR = [
            Port("B_BR_{}".format(i), (d + i * sep, -a), 0.5, 270) for i in range(N)
        ]
        ports_B_BL = [
            Port("B_BL_{}".format(i), (-d - i * sep, -a), 0.5, 270) for i in range(N)
        ]

        ports_B = [ports_B_TR, ports_B_TL, ports_B_BR, ports_B_BL]

    if config in ["A", "C"]:
        for ports1, ports2 in zip(ports_A, ports_B):
            elements = connect_bundle(ports1, ports2)
            top_cell.add(elements)

    elif config in ["B", "D"]:
        for ports1, ports2 in zip(ports_A, ports_B):
            elements = connect_bundle(ports2, ports1)
            top_cell.add(elements)

    return top_cell


@pp.autoname
def test_connect_bundle_udirect(dy=200, angle=270):

    xs1 = [-100, -90, -80, -55, -35, 24, 0] + [200, 210, 240]

    axis = "X" if angle in [0, 180] else "Y"

    pitch = 10.0
    N = len(xs1)
    xs2 = [50 + i * pitch for i in range(N)]

    if axis == "X":
        ports1 = [Port("top_{}".format(i), (0, xs1[i]), 0.5, angle) for i in range(N)]

        ports2 = [
            Port("bottom_{}".format(i), (dy, xs2[i]), 0.5, angle) for i in range(N)
        ]

    else:
        ports1 = [Port("top_{}".format(i), (xs1[i], 0), 0.5, angle) for i in range(N)]

        ports2 = [
            Port("bottom_{}".format(i), (xs2[i], dy), 0.5, angle) for i in range(N)
        ]

    top_cell = pp.Component(name="connect_bundle")
    elements = connect_bundle(ports1, ports2)
    for e in elements:
        top_cell.add(e)

    return top_cell


@pp.autoname
def test_connect_bundle_u_indirect(dy=-200, angle=180):

    xs1 = [-100, -90, -80, -55, -35] + [200, 210, 240]

    axis = "X" if angle in [0, 180] else "Y"

    pitch = 10.0
    N = len(xs1)
    xs2 = [50 + i * pitch for i in range(N)]

    a1 = angle
    a2 = a1 + 180

    if axis == "X":
        ports1 = [Port("top_{}".format(i), (0, xs1[i]), 0.5, a1) for i in range(N)]

        ports2 = [Port("bottom_{}".format(i), (dy, xs2[i]), 0.5, a2) for i in range(N)]

    else:
        ports1 = [Port("top_{}".format(i), (xs1[i], 0), 0.5, a1) for i in range(N)]

        ports2 = [Port("bottom_{}".format(i), (xs2[i], dy), 0.5, a2) for i in range(N)]

    top_cell = pp.Component()
    elements = connect_bundle(ports1, ports2)
    for e in elements:
        top_cell.add(e)

    return top_cell


@pp.autoname
def test_facing_ports():

    dy = 200.0
    xs1 = [-500, -300, -100, -90, -80, -55, -35, 200, 210, 240, 500, 650]

    pitch = 10.0
    N = len(xs1)
    xs2 = [-20 + i * pitch for i in range(N // 2)]
    xs2 += [400 + i * pitch for i in range(N // 2)]

    a1 = 90
    a2 = a1 + 180

    ports1 = [Port("top_{}".format(i), (xs1[i], 0), 0.5, a1) for i in range(N)]
    ports2 = [Port("bottom_{}".format(i), (xs2[i], dy), 0.5, a2) for i in range(N)]

    top_cell = pp.Component()
    elements = connect_bundle(ports1, ports2)
    # elements = link_ports_path_length_match(ports1, ports2)
    top_cell.add(elements)

    return top_cell


if __name__ == "__main__":
    import pp

    c = test_facing_ports()
    pp.show(c)



import pathlib
from phidl import quickplot


from pp.import_gds import import_gds


def plotgds(
    gds,
    show_ports=True,
    show_subports=False,
    label_ports=True,
    label_aliases=False,
    new_window=False,
):
    if isinstance(gds, (str, pathlib.Path)):
        gds = import_gds(gds)

    quickplot(
        gds,
        show_ports=show_ports,
        label_ports=label_ports,
        label_aliases=label_aliases,
        new_window=new_window,
    )


if __name__ == "__main__":
    import pp

    c = pp.c.waveguide()
    plotgds(c)



from pp.import_gds import import_gds
from pp.write_component import write_gds
import os
from pp.config import CONFIG
import sys


def remove_cell(device, cell_name):
    all_cells = device.get_dependencies(recursive=True)
    all_cell_names = set([c.name for c in all_cells])
    if cell_name in all_cell_names:
        for c in all_cells:
            to_remove = []
            for e in c.references:
                to_remove += [e]
            for e in to_remove:
                # print("removing", e)
                c.remove(e)


def is_empty(c):
    # print(len(c.polygons), len(c.get_dependencies()))
    return len(c.polygons) == 0 and len(c.get_dependencies()) == 0


def remove_empty_cells(gds, recursive=True, recurse_depth=0):
    """
    returns the list of removed cells
    """
    if type(gds) == str:
        gds = import_gds(gds)

    cells = gds.get_dependencies(recursive=True)
    _to_remove = []
    for c in cells:
        if is_empty(c):
            _to_remove += [c]

    for c in _to_remove:
        print(c.name)
        remove_cell(gds, c.name)

    if recursive:
        while (len(_to_remove)) > 0:
            print(recurse_depth)
            sys.stdout.flush()
            _to_remove = remove_empty_cells(
                gds, recursive=True, recurse_depth=recurse_depth + 1
            )

    return _to_remove


def remove_empty_cells_from_gds_file(gdspath):
    gds = import_gds(gdspath)
    remove_empty_cells(gds)
    write_gds(gds, gdspath[:-4] + "_cleaned.gds")


def clean_teg1():
    path_teg1 = os.path.join(CONFIG["mask_directory"], "teg1.gds")
    remove_empty_cells_from_gds_file(path_teg1)


if __name__ == "__main__":
    clean_teg1()



import numpy as np
import pp
from pp.geo_utils import extrude_path
from pp.components.euler.geo_euler import euler_bend_points
from pp.components.euler.geo_euler import euler_length
from pp.layers import LAYER
from pp.port import auto_rename_ports
from pp.component import Component
from typing import Tuple, Union


def _bend_euler(
    theta: int = 90,
    radius: Union[int, float] = 10.0,
    width: float = 0.5,
    resolution: float = 150.0,
    layer: Tuple[int, int] = LAYER.WG,
) -> Component:
    c = pp.Component()
    backbone = euler_bend_points(theta, radius=radius, resolution=resolution)
    pts = extrude_path(backbone, width)

    c.add_polygon(pts, layer=layer)
    # print(backbone[0])
    # print(backbone[-1])
    c.info["length"] = euler_length(radius, theta)
    c.radius = radius
    c.add_port(
        name="in0",
        midpoint=np.round(backbone[0].xy, 3),
        orientation=180,
        layer=layer,
        width=width,
    )
    c.add_port(
        name="out0",
        midpoint=np.round(backbone[-1].xy, 3),
        orientation=theta,
        layer=layer,
        width=width,
    )

    return c


@pp.autoname
def bend_euler90(
    radius: Union[int, float] = 10.0,
    width: float = 0.5,
    resolution: float = 150.0,
    layer: Tuple[int, int] = LAYER.WG,
) -> Component:
    """
    .. plot::
      :include-source:

      import pp

      c = pp.c.bend_euler90()
      pp.plotgds(c)

    """
    c = _bend_euler(
        theta=90, radius=radius, width=width, resolution=resolution, layer=layer
    )
    return auto_rename_ports(c)


@pp.autoname
def bend_euler90_biased(radius=10.0, width=0.5, resolution=150.0, layer=LAYER.WG):
    width = pp.bias.width(width)
    c = _bend_euler(
        theta=90, radius=radius, width=width, resolution=resolution, layer=layer
    )
    return auto_rename_ports(c)


@pp.autoname
def bend_euler180(
    radius: Union[int, float] = 10.0,
    width: float = 0.5,
    resolution: float = 150.0,
    layer: Tuple[int, int] = LAYER.WG,
) -> Component:
    """
    .. plot::
      :include-source:

      import pp

      c = pp.c.bend_euler180()
      pp.plotgds(c)

    """
    c = _bend_euler(
        theta=180, radius=radius, width=width, resolution=resolution, layer=layer
    )
    return auto_rename_ports(c)


@pp.autoname
def bend_euler180_biased(radius=10.0, width=0.5, resolution=150.0, layer=LAYER.WG):
    width = pp.bias.width(width)
    c = _bend_euler(
        theta=180, radius=radius, width=width, resolution=resolution, layer=layer
    )
    return auto_rename_ports(c)


if __name__ == "__main__":
    pass
    c = bend_euler90()
    # c = bend_euler90_biased()
    # c = bend_euler180()
    print(c.ports)
    pp.show(c)



import numpy as np
import pp
from pp.routing.manhattan import remove_flat_angles
from pp.routing.manhattan import round_corners

from pp.components import bend_circular
from pp.components import waveguide
from pp.components import taper as taper_factory

from pp.config import WG_EXPANDED_WIDTH, TAPER_LENGTH

from pp.routing.utils import get_list_ports_angle


def _is_vertical(segment, tol=1e-5):
    p0, p1 = segment
    return abs(p0[0] - p1[0]) < tol


def _is_horizontal(segment, tol=1e-5):
    p0, p1 = segment
    return abs(p0[1] - p1[1]) < tol


def _segment_sign(s):
    p0, p1 = s
    if _is_vertical(s):
        return np.sign(p1[1] - p0[1])

    if _is_horizontal(s):
        return np.sign(p1[0] - p0[0])


def get_ports_x_or_y_distances(list_ports, ref_point):
    if not list_ports:
        return []

    angle = get_list_ports_angle(list_ports)
    x0 = ref_point[0]
    y0 = ref_point[1]
    if angle in [0, 180]:
        xys = [p.y - y0 for p in list_ports]
        xs = [round(p.x, 5) for p in list_ports]
        if len(set(xs)) > 1:
            raise ValueError("List ports with angle 0/180 should all have the same x")
    else:
        xys = [p.x - x0 for p in list_ports]
        ys = [round(p.y, 5) for p in list_ports]
        if len(set(ys)) > 1:
            raise ValueError("List ports with angle 90/270 should all have the same y")
    return xys


def _distance(port1, port2):
    if hasattr(port1, "x"):
        x1, y1 = port1.x, port1.y
    else:
        x1, y1 = port1[0], port1[1]

    if hasattr(port2, "x"):
        x2, y2 = port2.x, port2.y
    else:
        x2, y2 = port2[0], port2[1]

    dx = x1 - x2
    dy = y1 - y2

    return np.sqrt(dx ** 2 + dy ** 2)


def connect_bundle_waypoints(
    start_ports,
    end_ports,
    way_points,
    straight_factory=waveguide,
    taper_factory=taper_factory,
    bend_factory=bend_circular,
    bend_radius=10.0,
    auto_sort=True,
    **kwargs
):
    """
    Args:
        start_ports: list of ports
        end_ports: list of ports
        way_points: list of points defining a route

    """
    if len(end_ports) != len(start_ports):
        raise ValueError(
            "Number of start ports should match number of end ports.\
        Got {} {}".format(
                len(start_ports), len(end_ports)
            )
        )

    for p in start_ports:
        p.angle = int(p.angle) % 360

    for p in end_ports:
        p.angle = int(p.angle) % 360

    start_angle = start_ports[0].orientation
    end_angle = end_ports[0].orientation

    # Sort the ports such that the bundle connect the correct corresponding ports.

    angles_to_sorttypes = {
        (0, 180): ("Y", "Y"),
        (0, 90): ("Y", "X"),
        (0, 0): ("Y", "-Y"),
        (0, 270): ("Y", "-X"),
        (90, 0): ("X", "Y"),
        (90, 90): ("X", "-X"),
        (90, 180): ("X", "-Y"),
        (90, 270): ("X", "X"),
        (180, 90): ("Y", "-X"),
        (180, 0): ("Y", "Y"),
        (180, 270): ("Y", "X"),
        (180, 180): ("Y", "-Y"),
        (270, 90): ("X", "X"),
        (270, 270): ("X", "-X"),
        (270, 0): ("X", "-Y"),
        (270, 180): ("X", "Y"),
    }

    dict_sorts = {
        "X": lambda p: p.x,
        "Y": lambda p: p.y,
        "-X": lambda p: -p.x,
        "-Y": lambda p: -p.y,
    }
    key = (start_angle, end_angle)
    sp_st, ep_st = angles_to_sorttypes[key]
    start_port_sort = dict_sorts[sp_st]
    end_port_sort = dict_sorts[ep_st]

    if auto_sort:
        start_ports.sort(key=start_port_sort)
        end_ports.sort(key=end_port_sort)

    routes = _generate_manhattan_bundle_waypoints(
        start_ports, end_ports, way_points, **kwargs
    )

    bends90 = [bend_factory(radius=bend_radius, width=p.width) for p in start_ports]

    if taper_factory:
        if callable(taper_factory):
            taper = taper_factory(
                length=TAPER_LENGTH,
                width1=start_ports[0].width,
                width2=WG_EXPANDED_WIDTH,
                layer=start_ports[0].layer,
            )
        else:
            # In this case the taper is a fixed cell
            taper = taper_factory
    else:
        taper = None
    connections = [
        round_corners(pts, bend90, straight_factory, taper=taper)
        for pts, bend90 in zip(routes, bends90)
    ]
    return connections


def snap_route_to_end_point_x(route, x):
    y1, y2 = [p[1] for p in route[-2:]]
    return route[:-2] + [(x, y1), (x, y2)]


def snap_route_to_end_point_y(route, y):
    x1, x2 = [p[0] for p in route[-2:]]
    return route[:-2] + [(x1, y), (x2, y)]


def _generate_manhattan_bundle_waypoints(
    start_ports, end_ports, backbone_route, **kwargs
):
    """
    Args:
        start_ports: list of start ports. Should all be facing in the same direction
        end_ports: list of end ports. Should all be facing in the same direction
        route: going from one point somewhere within the start_ports bank to
        another point within the end_ports bank
    """
    backbone_route = remove_flat_angles(backbone_route)

    way_segments = [(p0, p1) for p0, p1 in zip(backbone_route[:-1], backbone_route[1:])]
    offsets_start = get_ports_x_or_y_distances(start_ports, backbone_route[0])

    start_angle = start_ports[0].orientation
    if start_angle in [90, 270]:
        offsets_start = [-_d for _d in offsets_start]

    end_angle = end_ports[0].orientation

    def _displace_segment_copy(s, a, sh=1, sv=1):
        sign_seg = _segment_sign(s)
        if _is_horizontal(s):
            dp = (0, sh * sign_seg * a)
        elif _is_vertical(s):
            dp = (sv * sign_seg * a, 0)
        else:
            raise ValueError("Segment should be manhattan, got {}".format(s))

        displaced_seg = [np.array(p) + dp for p in s]
        return displaced_seg

    def _displace_segment_copy_group1(s, a):
        return _displace_segment_copy(s, a, sh=1, sv=-1)

    def _intersection(s1, s2):
        if _is_horizontal(s1) and _is_vertical(s2):
            sh, sv = s1, s2
        elif _is_horizontal(s2) and _is_vertical(s1):
            sh, sv = s2, s1
        else:

            if _is_horizontal(s1):
                s1_dir = "h"
            elif _is_vertical(s1):
                s1_dir = "v"
            else:
                s1_dir = "u"

            if _is_horizontal(s2):
                s2_dir = "h"
            elif _is_vertical(s2):
                s2_dir = "v"
            else:
                s2_dir = "u"

            raise ValueError(
                "s1 / s2 should be h/v or v/h. Got \
            {} {} {} {}".format(
                    s1_dir, s2_dir, s1, s2
                )
            )
        return (sv[0][0], sh[0][1])

    N = len(start_ports)
    routes = []
    _make_segment = _displace_segment_copy_group1
    for i in range(N):
        prev_seg_sep = offsets_start[i]
        route = []

        for j, seg in enumerate(way_segments):
            seg_sep = prev_seg_sep
            d_seg = _make_segment(seg, seg_sep)

            if j == 0:
                start_point = d_seg[0]
            else:
                prev_seg = way_segments[j - 1]
                tmp_seg = _make_segment(prev_seg, prev_seg_sep)
                start_point = _intersection(d_seg, tmp_seg)

            route += [start_point]

            # If last point before the ports, adjust the separation to the end ports
            if j == len(way_segments) - 1:
                end_point = end_ports[i].position
                route += [end_point]

                if end_angle in [0, 180]:
                    route = snap_route_to_end_point_y(route, end_point[1])
                else:
                    route = snap_route_to_end_point_x(route, end_point[0])

        routes += [route]

    return routes


@pp.autoname
def test_connect_bundle_waypoints():
    return test_connect_bundle_waypointsA()


@pp.autoname
def test_connect_bundle_waypointsA():
    import pp
    from pp.component import Port

    xs1 = np.arange(10) * 5 - 500.0

    N = xs1.size
    ys2 = np.array([0, 5, 10, 20, 25, 30, 40, 55, 60, 75]) + 500.0

    ports1 = [Port("A_{}".format(i), (xs1[i], 0), 0.5, 90) for i in range(N)]
    ports2 = [Port("B_{}".format(i), (0, ys2[i]), 0.5, 180) for i in range(N)]

    top_cell = pp.Component()
    p0 = ports1[0].position + (22, 0)
    way_points = [
        p0,
        p0 + (0, 100),
        p0 + (200, 100),
        p0 + (200, -200),
        p0 + (0, -200),
        p0 + (0, -350),
        p0 + (400, -350),
        (p0[0] + 400, ports2[-1].y),
        ports2[-1].position,
    ]

    elements = connect_bundle_waypoints(ports1, ports2, way_points)
    top_cell.add(elements)

    return top_cell


@pp.autoname
def test_connect_bundle_waypointsB():
    import pp
    from pp.component import Port

    ys1 = np.array([0, 5, 10, 15, 30, 40, 50, 60]) + 0.0
    ys2 = np.array([0, 10, 20, 30, 70, 90, 110, 120]) + 500.0
    N = ys1.size

    ports1 = [Port("A_{}".format(i), (0, ys1[i]), 0.5, 0) for i in range(N)]
    ports2 = [Port("B_{}".format(i), (500, ys2[i]), 0.5, 180) for i in range(N)]

    p0 = ports1[0].position + (0, 22.5)

    top_cell = pp.Component()
    way_points = [
        p0,
        p0 + (200, 0),
        p0 + (200, -200),
        p0 + (400, -200),
        (p0[0] + 400, ports2[0].y),
        ports2[0].position,
    ]

    elements = connect_bundle_waypoints(ports1, ports2, way_points)
    top_cell.add(elements)

    return top_cell


@pp.autoname
def test_connect_bundle_waypointsC():
    import pp
    from pp.component import Port

    ys1 = np.array([0, 5, 10, 15, 20, 60, 70, 80, 120, 125])
    ys2 = np.array([0, 5, 10, 20, 25, 30, 40, 55, 60, 65]) - 500.0
    N = ys1.size

    ports1 = [Port("A_{}".format(i), (0, ys1[i]), 0.5, 0) for i in range(N)]
    ports2 = [Port("B_{}".format(i), (600, ys2[i]), 0.5, 180) for i in range(N)]

    top_cell = pp.Component()
    way_points = [
        ports1[0].position,
        ports1[0].position + (200, 0),
        ports1[0].position + (200, -200),
        ports1[0].position + (400, -200),
        (ports1[0].x + 400, ports2[0].y),
        ports2[0].position,
    ]

    elements = connect_bundle_waypoints(ports1, ports2, way_points)
    top_cell.add(elements)

    return top_cell


@pp.autoname
def test_connect_bundle_waypointsD():
    import pp
    from pp.component import Port

    ys1 = np.array([0, 5, 10, 20, 25, 30, 40, 55, 60, 75]) + 100.0
    ys2 = np.array([0, -5, -10, -20, -25, -30, -40, -55, -60, -75]) + 500.0
    N = ys1.size

    ports1 = [Port("A_{}".format(i), (0, ys1[i]), 0.5, 0) for i in range(N)]
    ports2 = [Port("B_{}".format(i), (0, ys2[i]), 0.5, 0) for i in range(N)]

    def _mean_y(ports):
        return np.mean([p.y for p in ports])

    yc1 = _mean_y(ports1)
    yc2 = _mean_y(ports2)

    top_cell = pp.Component()
    way_points = [(0, yc1), (200, yc1), (200, yc2), (0, yc2)]

    elements = connect_bundle_waypoints(ports1, ports2, way_points)
    top_cell.add(elements)

    return top_cell


if __name__ == "__main__":
    cell = test_connect_bundle_waypointsD()
    pp.show(cell)



from pp.name import autoname
from pp.component import Component
from pp.layers import LAYER
from pp.port import deco_rename_ports
from pp.components.hline import hline


WIRE_WIDTH = 10.0


@deco_rename_ports
@autoname
def wire(length=50.0, width=WIRE_WIDTH, layer=LAYER.M3):
    """ electrical straight wire

    .. plot::
      :include-source:

      import pp

      c = pp.c.wire(length=50., width=10., layer=pp.LAYER.M3)
      pp.plotgds(c)
    """
    return hline(length=length, width=width, layer=layer)


@deco_rename_ports
@autoname
def corner(width=WIRE_WIDTH, radius=None, layer=LAYER.M3):
    """ 90 degrees electrical bend

    Args:
        width: wire width
        radius ignore (passed for consistency with other types of bends)
        layer: layer

    .. plot::
      :include-source:

      import pp

      c = pp.c.corner(width=10., layer=pp.LAYER.M3)
      pp.plotgds(c)
    """
    c = Component()
    a = width / 2
    xpts = [-a, a, a, -a]
    ypts = [-a, -a, a, a]

    c.add_polygon([xpts, ypts], layer=layer)
    c.add_port(name="W0", midpoint=(-a, 0), width=width, orientation=180)
    c.add_port(name="N0", midpoint=(0, a), width=width, orientation=90)
    c.info["length"] = width
    return c


if __name__ == "__main__":
    import pp

    c = wire()
    pp.show(c)



""" Command line interface for gdsfactory
"""

import os
import re
import shlex
import subprocess
import time
import pathlib
import click
import git

from pp import CONFIG
from pp.layers import LAYER
from pp.config import logging
from pp import klive
from pp.config import print_config

# from pp.write_doe_from_yaml import write_doe_from_yaml
from pp.write_doe_from_yaml import import_custom_doe_factories

from pp.mask.merge_json import merge_json
from pp.mask.merge_markdown import merge_markdown
from pp.mask.merge_test_metadata import merge_test_metadata
from pp.mask.write_labels import write_labels

import pp.build as pb

from pp.tests.test_factory import lock_components_with_changes


VERSION = "1.4.2"
log_directory = CONFIG.get("log_directory")
cwd = pathlib.Path.cwd()
LAYER_LABEL = LAYER.LABEL


def shorten_command(cmd):
    """ Shortens a command if possible"""
    match = re.search(r"(\w+\.(py|drc))", cmd)
    return match.group() if match else cmd


def run_command(command):
    """ Run a command and keep track of some context """
    logging.info("Running `{}`".format(command))

    # Run the process and handle errors
    time0 = time.time()
    process = subprocess.Popen(
        shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = process.communicate()
    total_time = time.time() - time0

    # Either show that ther was an error, or just leave it
    if process.returncode == 0:
        message = "`{}` ran without errors in {:.2f}s.".format(
            shorten_command(command), total_time
        )
        logging.info(message)
        if stdout.strip():
            message = "Output of `{}`:".format(shorten_command(command))
            logging.info(message)
            logging.info(stdout.strip(), extra={"raw": True})
    else:
        message = "Error in `{}`".format(shorten_command(command))
        logging.error(message)
        raw = stdout.strip() + "\n" + stderr.strip()
        logging.error(raw, extra={"raw": True})

    return command, process.returncode


def print_version(ctx, param, value):
    """ Prints the version """
    if not value or ctx.resilient_parsing:
        return
    click.echo(VERSION)
    ctx.exit()


"""
LOGS
"""


@click.group()
def log():
    """ Work with logs """
    pass


@click.command(name="show")
@click.argument("logfile", default="main", required=False)
def log_show(logfile):
    """ Show logs """
    if not os.path.exists(log_directory):
        print("No logs found.")
        return

    filename = os.path.join(log_directory, "{}.log".format(logfile))
    with open(filename) as f:
        data = f.read()
        print(data.strip() if data else "Logs are empty")


@click.command(name="delete")
@click.argument("logfile", default="main", required=False)
def log_delete(logfile):
    """ Deletes logs """
    if not os.path.exists(log_directory):
        print("No logs found.")
        return

    filename = os.path.join(log_directory, "{}.log".format(logfile))
    subprocess.check_output(["rm", filename])


"""
CONFIG
"""


@click.command(name="config")
@click.argument("key", required=False, default=None)
def config_get(key):
    """ Shows key values from CONFIG """
    print_config(key)


"""
LIBRARY
"""


@click.group()
def library():
    """ Commands for managing libraries """
    pass


@click.argument("dirname", required=False, default=None)
@click.command(name="pull")
def library_pull(dirname):
    """ Pull the library repo """
    if dirname:
        repo_path = cwd / dirname
    else:
        repo_path = CONFIG["gdslib"]

    if os.path.isdir(repo_path):
        print("git pull: {}".format(repo_path))
        g = git.cmd.Git(repo_path)
        g.pull()


@click.command(name="lock")
@click.argument("dirname", required=False, default=None)
def library_lock(dirname):
    """ lock component with changes """

    if dirname:
        path_library = cwd / dirname
    else:
        path_library = CONFIG["gdslib"]

    if os.path.isdir(path_library):
        lock_components_with_changes(path_library=path_library)


"""
MASKS
"""


@click.group()
def mask():
    """ Commands for building masks """
    pass


@click.command(name="clean")
@click.option("--force", "-f", default=False, help="Force deletion", is_flag=True)
def build_clean(force):
    """ Deletes the build folder and contents """
    message = "Delete {}. Are you sure?".format(CONFIG["build_directory"])
    if force or click.confirm(message, default=True):
        pb.build_clean()


@click.command(name="build_devices")
@click.argument("regex", required=False, default=".*")
def build_devices(regex):
    """ Build all devices described in devices/"""
    pb.build_devices(regex)


@click.command(name="build_does")
def build_does():
    """ Build does defined in doe.yml"""
    print("this is deprecated")
    import_custom_doe_factories()
    # write_doe_from_yaml()
    pb.build_does()


@click.command(name="write_metadata")
@click.argument("label_layer", required=False, default=LAYER_LABEL)
def mask_merge(label_layer):
    """ merge JSON/Markdown from build/devices into build/mask"""

    gdspath = CONFIG["mask_gds"]
    write_labels(gdspath=gdspath, label_layer=label_layer)

    merge_json()
    merge_markdown()
    merge_test_metadata(config_path=CONFIG["config_path"])


@click.command(name="write_labels")
@click.argument("gdspath", default=None)
@click.argument("label_layer", required=False, default=LAYER_LABEL)
def write_mask_labels(gdspath, label_layer):
    """ find test and measurement labels """
    if gdspath is None:
        gdspath = CONFIG["mask_gds"]

    write_labels(gdspath=gdspath, label_layer=label_layer)


"""
EXTRA
"""


@click.command()
@click.argument("filename")
def show(filename):
    """ Show a GDS file using KLive """
    klive.show(filename)


@click.command()
def test():
    """ Run tests using pytest.
    Strictly speaking you should just run `pytest` directly."""

    os.chdir(CONFIG["repo_path"])
    command = shlex.split("pytest")
    subprocess.call(command)


@click.group()
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Show the version number.",
)
def cli():
    """ `pf` is the photonics factory command line tool.
    It helps to build, test, and configure masks and components.
    """
    pass


log.add_command(log_show)
log.add_command(log_delete)

library.add_command(library_pull)
library.add_command(library_lock)

mask.add_command(build_clean)
mask.add_command(build_devices)
mask.add_command(build_does)
mask.add_command(mask_merge)
mask.add_command(write_mask_labels)

cli.add_command(config_get)
cli.add_command(library)
cli.add_command(log)
cli.add_command(mask)
cli.add_command(show)
cli.add_command(test)


if __name__ == "__main__":
    cli()



import pp
from pp.components.bend_s import bend_s
from pp.component import Component
from typing import Callable, List, Tuple


@pp.autoname
def coupler_symmetric(
    bend: Callable = bend_s,
    gap: float = 0.234,
    wg_width: float = 0.5,
    layer: Tuple[int, int] = pp.LAYER.WG,
    layers_cladding: List[Tuple[int, int]] = [pp.LAYER.WGCLAD],
    cladding_offset: float = 3.0,
) -> Component:
    """ two coupled waveguides with bends

    Args:
        bend:
        gap: um

    .. plot::
      :include-source:

      import pp

      c = pp.c.coupler_symmetric()
      pp.plotgds(c)

    """
    bend = pp.call_if_func(
        bend,
        width=wg_width,
        layer=layer,
        layers_cladding=layers_cladding,
        cladding_offset=cladding_offset,
        pins=False,
    )

    w = bend.ports["W0"].width
    y = (w + gap) / 2

    c = pp.Component()
    top_bend = bend.ref(position=(0, y), port_id="W0")
    bottom_bend = bend.ref(position=(0, -y), port_id="W0", v_mirror=True)

    c.add(top_bend)
    c.add(bottom_bend)

    # Using absorb here to have a flat cell and avoid
    # to have deeper hierarchy than needed
    c.absorb(top_bend)
    c.absorb(bottom_bend)

    port_width = 2 * w + gap
    c.add_port(name="W0", midpoint=[0, 0], width=port_width, orientation=180)
    c.add_port(port=bottom_bend.ports["E0"], name="E0")
    c.add_port(port=top_bend.ports["E0"], name="E1")

    return c


@pp.autoname
def coupler_symmetric_biased(bend=bend_s, gap=0.2, wg_width=0.5):
    return coupler_symmetric(
        bend=bend, gap=pp.bias.gap(gap), wg_width=pp.bias.width(wg_width)
    )


def _demo_coupler_symmetric(wg_width=0.5, gap=0.2):
    # c = coupler_symmetric(gap=gap, wg_width=wg_width)
    c = coupler_symmetric_biased(gap=gap, wg_width=wg_width)
    pp.write_gds(c)
    return c


if __name__ == "__main__":
    c = _demo_coupler_symmetric()
    pp.show(c)



import picwriter.components as pc

from pp.components.waveguide_template import wg_strip
from pp.picwriter2component import picwriter2component
import pp
from pp.component import Component
from typing import Callable, Tuple


@pp.autoname
def disk(
    radius: float = 10.0,
    gap: float = 0.2,
    wrap_angle: int = 0,
    parity: int = 1,
    port: Tuple[int, int] = (0, 0),
    direction: str = "EAST",
    waveguide_template: Callable = wg_strip,
    **kwargs
) -> Component:
    """ Disk Resonator

    Args:
       radius (float): Radius of the disk resonator
       gap (float): Distance between the bus waveguide and resonator
       wrap_angle (float): Angle in *radians* between 0 and pi (defaults to 0) that determines how much the bus waveguide wraps along the resonator.  0 corresponds to a straight bus waveguide, and pi corresponds to a bus waveguide wrapped around half of the resonator.
       parity (1 or -1): If 1, resonator to left of bus waveguide, if -1 resonator to the right
       port (tuple): Cartesian coordinate of the input port (x1, y1)
       direction (string): Direction that the component will point *towards*, can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`, OR an angle (float, in radians)
       waveguide_template (WaveguideTemplate): Picwriter WaveguideTemplate object

    Where in the above (x1,y1) is the same as the 'port' input, (x2, y2) is the end of the component, and 'dir1', 'dir2' are of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`, *or* an angle in *radians*.
    'Direction' points *towards* the waveguide that will connect to it.

    Other Parameters:
       wg_width: 0.5
       wg_layer: pp.LAYER.WG[0]
       wg_datatype: pp.LAYER.WG[1]
       clad_layer: pp.LAYER.WGCLAD[0]
       clad_datatype: pp.LAYER.WGCLAD[1]
       bend_radius: 10
       cladding_offset: 3

    .. plot::
      :include-source:

      import pp

      c = pp.c.disk(radius=10, wrap_angle=3.14/4)
      pp.plotgds(c)

    """

    c = pc.Disk(
        pp.call_if_func(wg_strip, **kwargs),
        radius=radius,
        coupling_gap=gap,
        wrap_angle=wrap_angle,
        parity=parity,
        port=port,
        direction=direction,
    )

    return picwriter2component(c)


if __name__ == "__main__":
    import pp

    c = disk(wrap_angle=3.14 / 4, pins=True)
    pp.show(c)



import hashlib
import pathlib
import git
import time
import os
import pp
from pp.config import CONFIG, conf


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def list_gds_files(gdslib=CONFIG["gdslib"], tech=conf.tech.name):
    gdslib_path = pathlib.Path(gdslib) / tech
    return [gds for gds in gdslib_path.glob("*.gds")]


def pull_library(repo_path=CONFIG["gdslib"]):
    if os.path.isdir(repo_path):
        print("git pull: {}".format(repo_path))
        g = git.cmd.Git(repo_path)
        g.pull()


def load_library(gdslib=CONFIG["gdslib"], tech=conf.tech.name):
    """
    Load a cell library from the tech
    """
    lib = {}
    gds_files = list_gds_files(gdslib, tech)

    for filepath in gds_files:
        filename = filepath.stem
        lib[filename] = pp.load_component(
            filename, filepath.parent, with_info_labels=False
        )

    return lib


def print_lib_info(gdslib=CONFIG["gdslib"], dump_file=None):
    """
    cell    cell_path   md5_hash
    """
    gds_files = list_gds_files(gdslib)
    if not gds_files:
        return

    names = []
    filepaths = []
    geohashes = []
    build_dates = []
    for filepath in gds_files:
        root, name = os.path.split(filepath)
        name = name[:-4]
        geohash = pp.load_component(name, root).hash_geometry()
        names += [name]
        filepaths += [str(filepath)]
        geohashes += [geohash]
        build_dates += [time.ctime(os.path.getmtime(filepath))]

    n_name = max([len(x) for x in names]) + 1
    n_path = max([len(x) for x in filepaths]) + 1
    n_geohash = max([len(x) for x in geohashes]) + 1
    n_build_date = max([len(x) for x in build_dates]) + 1

    lines = []
    line = (
        "name".ljust(n_name)
        + "path".ljust(n_path)
        + "hash geometry".ljust(n_geohash)
        + "build date".ljust(n_build_date)
    )  # +'md5 hash'.ljust(n_hash)
    lines += [line]
    lines += ["-" * len(line)]
    for name, path, geohash, build_date in zip(
        names, filepaths, geohashes, build_dates
    ):
        line = "{}{}{}{}".format(
            name.ljust(n_name),
            path.ljust(n_path),
            geohash.ljust(n_geohash),
            build_date.ljust(n_build_date),
        )
        lines += [line]

    assert not dump_file.endswith(".gds")
    if dump_file is None:
        print()
        for line in lines:
            print(line)
        print()

    elif type(dump_file) == str:
        with open(dump_file, "w") as fw:
            for line in lines:
                fw.write(line + "\n")

    elif hasattr(dump_file, "write"):
        for line in lines:
            dump_file.write(line + "\n")
    else:
        raise ValueError("invalid filepath, {}".format(dump_file))


if __name__ == "__main__":
    print(CONFIG["gdslib"])
    print(list_gds_files())
    lib = load_library()
    print(lib.keys())
